/**h* NFCTools/nfc_create_main.c
 *
 * NAME
 *   nfc_create_main.c
 *
 * DESCRIPTION
 *   Main of nfc_create
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"

#ifdef WIN32
#include "getopt/getopt.h"
#include "getopt/getopt_long.h"
#endif

#ifdef __linux__
#include <getopt.h>
#endif

static void hello(void);
static void usage(void);
static void helpmsg(const char *str);

BOOL nfc_verbose = FALSE;
BOOL nfc_silent  = FALSE;

const char prog_name[] = "nfccreate";
const char short_options[] = "u:t:a:m:z:i:o:bcr:lpkvqhy";

const struct option long_options[] = {

   { "uri",            1, NULL, 'u' },
   { "title",          1, NULL, 't' },
	 { "action",         1, NULL, 'a' },
	 { "mime-type",      1, NULL, 'm' },
	 { "size",           1, NULL, 'z' },

	 { "input-file",     1, NULL, 'i' },
	 { "output-file",    1, NULL, 'o' },
	 { "overwrite-file", 0, NULL, 0xA0 },

	 { "batch",          0, NULL, 'b' },
	 { "continue",       0, NULL, 'c' },

	 { "reader",         1, NULL, 'r' },
	 { "list-readers",   1, NULL, 'l' },
	 { "prepare",        0, NULL, 'p' },
	 { "lock",           0, NULL, 'k' },
	 { "overwrite-tag",  0, NULL, 0xA1 },

	 { "verbose",        0, NULL, 'v' },
	 { "quiet",          0, NULL, 'q' },

	 { "help",           1, NULL, 'h' },

	 { "title-text",     1, NULL, 0x80 },
	 { "title-lang",     1, NULL, 0x81 },

	 { "icon-file",      1, NULL, 0x90 },
	 { "icon-type",      1, NULL, 0x91 },

};


int main(int argc, char **argv)
{
  char *p_uri = NULL;
  int p_title = 0;
  char *p_action = NULL;
  char *p_mime_type = NULL;
  unsigned long p_size = 0;
  char *p_icon_file = NULL;
	char *p_icon_type = NULL;
  char *p_reader = NULL;
	int p_prepare = 0;
  int p_lock = 0;
  char *p_input = NULL;
  char *p_output = NULL;
  int p_list = 0;
  int p_batch = 0;
  int p_continue = 0;
  int p_overwrite_tag = 0;
	int p_overwrite_file = 0;

	int c;

  nfc_tag_t tag = NULL;
	nfc_card_t card = NULL;
	unsigned char *ndef_buffer = NULL;
	size_t ndef_size;
	size_t ndef_tlv_size;

  while ((c = getopt_long(argc, argv, short_options, long_options, NULL)) != -1)
  {
    switch (c)
    {
      case 'u'  : p_uri = optarg; break;

      case 't'  : 
			case 0x80 : p_title = TRUE; break;
			case 0x81 : break;

      case 'a'  : p_action = optarg; break;
      case 'm'  : p_mime_type = optarg; break;
      case 'z'  : p_size = atol(optarg); break;

      case 'i'  : p_input = optarg; break;
      case 'o'  : p_output = optarg; break;

      case 'b'  : p_batch = 1; break;
      case 'c'  : p_continue = 1; break;

      case 'r'  : p_reader = optarg; break;
      case 'l'  : p_list = 1; break;
			case 'p'  : p_prepare = 1; break;
			case 'k'  : p_lock = 1; break;

      case 'v'  : nfc_verbose = TRUE; break;
      case 'q'  : nfc_silent = TRUE; break;

      case 'y'  : p_overwrite_file = p_overwrite_tag = 1; break;

      case 0x90 : p_icon_file = optarg; break;
      case 0x91 : p_icon_type = optarg; break;

			case 0xA0 : p_overwrite_file = 1; break;
			case 0xA1 : p_overwrite_tag = 1; break;

      case 'h'  : hello();
                  usage();
                  return EXIT_FAILURE;

      case '?'  : hello();
			            helpmsg("Incorrect parameter(s).\n");
                  return EXIT_FAILURE;
									break;

      default   : hello();
			            helpmsg("Invalid command line.\n");
                  return EXIT_FAILURE;
    }
  }

	if (!nfc_silent)
	  hello();

  if (p_list)
	{
    nfc_list_readers();
		return EXIT_SUCCESS;
	}

  /* Check the minimum mandatory parameters */
  if ((p_uri == NULL) && (p_input == NULL))
  {
    helpmsg("No URI parameter, and no Input file, nothing to do!\n");
    return EXIT_FAILURE;
  }
	if ((p_icon_type != NULL) && (p_icon_file == NULL))
	{
    helpmsg("Incoherent parameters for Icon.\n");
    return EXIT_FAILURE;
	}
  if ((p_reader == NULL) && (p_output == NULL))
  {
    helpmsg("No Reader selected and no Output file, nothing to do!\n");
    return EXIT_FAILURE;
  }

	/* Verify the reader */
	if (p_reader != NULL)
	{
	  p_reader = nfc_reader_find(p_reader);
		if (p_reader == NULL)
		{
      printf("Reader not found!\n");
      return EXIT_FAILURE;
		}
	}

  /* Create the NFC Tag object */
  tag = nfc_tac_alloc();
  if (tag == NULL)
    return EXIT_FAILURE;

  /* From now on, we shall not exit directly, but jump to label */
  /* 'failed' to have the allocated data freed before returning */
  if (p_input != NULL)
    p_input = nfc_file_name_alloc(p_input, "ndef");
  if (p_output != NULL)
    p_output = nfc_file_name_alloc(p_output, "ndef");

  /* Try to load the NFC Tag from the input file */
  if (p_input != NULL)
  {
    if (!nfc_file_exists(p_input))
    {
      printf("Input file doesn't exist.\n");
      goto failed;    
    }
    if (!nfc_tag_load_from_file(tag, p_input))
    {
      printf("Failed to load the Content of the NFC Tag.\n");
      goto failed;
    }  
  } else
  {
    if (!nfc_smartposter_init(tag))
    {
      printf("Failed to initialize the Content of the NFC Tag.\n");
      goto failed;
    }
  }

  /* Assign the new value(s) */
	/* ----------------------- */

	/* URI */
  if (p_uri != NULL)
  {
    if (!nfc_smartposter_set_uri(tag, p_uri))
    {
      printf("Failed to assign the URI.\n");
      goto failed;
    }
  }

  /* Title(s) */
  if (p_title)
  {
	  char *optalloc = NULL;
		char *lang = NULL;
		char *text = NULL;

    optind = 1;

		while ((c = getopt_long(argc, argv, short_options, long_options, NULL)) != -1)
		{
			switch (c)
			{
				case 't'  : if (strchr(optarg, ':'))
				            {
										  optalloc = nfc_strdup(optarg);
											lang = optalloc;
											text = strchr(optalloc, ':');
											*text = '\0';
											text ++;
										} else
										{
                      lang = NULL;
											text = optarg;
										}
										break;

				case 0x80 : text = optarg; break;
				case 0x81 : lang = optarg; break;

				default   : break;
			}

			if (text != NULL)
			{
				if (!nfc_smartposter_set_title(tag, text, lang))
				{
					printf("Failed to assign the Title.\n");
					goto failed;
				}
				text = NULL;
				lang = NULL;
			}
			if (optalloc != NULL)
			{
        nfc_free(optalloc);
				optalloc = NULL;
			}
		}
  }

	/* Action */
  if (p_action != NULL)
  {
	  int action = -1;

		if (!strcmp(optarg, "open") || !strcmp(optarg, "do"))
		{
			action = 0;
		} else
		if (!strcmp(optarg, "save"))
		{
			action = 1;
		} else
		if (!strcmp(optarg, "edit"))
		{
			action = 2;
		} else                 
		{
			action = atoi(optarg);
		}
		
		if (action < 0)
		{
      printf("Invalid Action.\n");
      goto failed;
		}

    if (!nfc_smartposter_set_action(tag, action))
    {
      printf("Failed to assign the Action.\n");
      goto failed;
    }
  }

  /* Size of target */
  if (p_size)
  {
    if (!nfc_smartposter_set_size(tag, p_size))
    {
      printf("Failed to assign the Size.\n");
      goto failed;
    }
  }

  /* Type of target */
  if (p_mime_type != NULL)
  {
    if (!nfc_smartposter_set_mime_type(tag, p_mime_type))
    {
      printf("Failed to assign the MIME-Type.\n");
      goto failed;
    }
  }

  /* Icon of the target */
  if (p_icon_file != NULL)
  {
    if (!nfc_file_exists(p_icon_file))
    {
      printf("Icon file doesn't exist.\n");
      goto failed;
    }
    if (nfc_file_size(p_icon_file) > MAX_NFC_TAG_SIZE/2)
    {
      printf("Icon file is too big.\n");
      goto failed;
    }
    if (!nfc_smartposter_load_icon(tag, p_icon_file))
    {
      printf("Failed to load the Icon.\n");
      goto failed;
    }

		if (p_icon_type == NULL)
		{
		  /* Guess the type of the icon */
		  p_icon_type = nfc_file_type_alloc(p_icon_file);
			if (p_icon_type == NULL)
			{
			  if (!nfc_silent)
				  printf("Type of icon is unknown\n");
			} else
			{
			  if (!nfc_smartposter_set_icon_mime_type(tag, p_icon_type))
				{
				  nfc_free(p_icon_type);
				  printf("Failed to assign the Icon MIME-Type.\n");
				  goto failed;
				}
				nfc_free(p_icon_type);
			}

		} else
		if (p_icon_type != NULL)
		{
		  /* Assign the type of the icon */
			if (!nfc_smartposter_set_icon_mime_type(tag, p_icon_type))
			{
				printf("Failed to assign the Icon MIME-Type.\n");
				goto failed;
			}
		}
	}

  /* Create the output */
	/* ----------------- */

	if (!ndef_assemble_tag_alloc(tag, &ndef_buffer, &ndef_size))
	{
		printf("Can't assemble the NDEF output\n");
		goto failed;
	}

	/* What will the final size be ? */
	if (ndef_size < 0xFF)
	  ndef_tlv_size = 2 + ndef_size;
	else
	  ndef_tlv_size = 4 + ndef_size;

  /* Save the (new) content to a (new) file */  
  if (p_output != NULL)
  {
    if (!p_overwrite_file && nfc_file_exists(p_output))
    {
      printf("Output file already exists. Try flag --overwrite-file.");
      goto failed;
    }
		if (!nfc_write_file(ndef_buffer, ndef_size, p_output))
		{
			printf("Failed to write the output file\n");
			goto failed;
		}
    if (!nfc_silent)
		  printf("NFC Tag Content saved.\n");
  }

  /* Write the (new) content to a (new) tag */
	/* -------------------------------------- */

	if (p_reader != NULL)
	{

		if (!nfc_silent)
			printf("Writing NFC Tag to reader '%s'\n", p_reader);

batch_begin:

		if (card != NULL)
		{
      nfc_card_close(card);
      card = NULL;
		}

    /* Wait for a card */
    if (p_batch)
		{
		  if (!nfc_silent)
			  printf("Put next NFC Tag on reader...\n");
			if (!nfc_reader_wait_card_present(p_reader))
			{
				printf("Failed to wait for card arrival.\n");
				goto failed;
			}
		} else
		{
		  if (!nfc_reader_is_card_present(p_reader))
			{
			  printf("Put a blank NFC Tag on the reader...\n");
				if (!nfc_reader_wait_card_present(p_reader))
				{
					printf("Failed to wait for card arrival.\n");
					goto failed;
				}
			}
		}

    /* Connect to the card */
    card = nfc_card_open(p_reader);
		if (card == NULL)
		{
      printf("Failed to connect to the target Tag.\n");
			if (p_batch && p_continue)
			  goto batch_begin;
		  goto failed;
		}

    /* Do whatever must be done to know what the card has in its womb */
		if (!nfc_card_recognize(card))
		{
      printf("Failed to detect the type of the connected Tag.\n");
			if (p_batch && p_continue)
			  goto batch_begin;
		  goto failed;
		}

    /* Check the type */
		switch (nfc_card_get_type(card))
		{
      case CARD_INNOVISION_TOPAZ   : if (!nfc_silent)
			                                 printf("Found a Type 1 Tag (Innovision Topaz)\n");
																		 break;
      case CARD_NXP_MIFARE_UL      : if (!nfc_silent)
			                                 printf("Found a Type 2 Tag (NXP Mifare UltraLight)\n");
																		 break;
      case CARD_NXP_MIFARE_UL_C    : if (!nfc_silent)
			                                 printf("Found a Type 2 Tag (NXP Mifare UltraLight C)\n");
																		 break;
      case CARD_NXP_MIFARE_1K      : if (!nfc_silent)
			                                 printf("Found a 'unofficial' Type 2 Tag (NXP Mifare Classic 1K)\n");
																		 break;
      case CARD_NXP_MIFARE_4K      : if (!nfc_silent)
			                                 printf("Found a 'unofficial' Type 2 Tag (NXP Mifare Classic 4K)\n");
																		 break;
      case CARD_NXP_DESFIRE_EV1    : if (!nfc_silent)
			                                 printf("Found a Type 4 Tag (NXP Desfire EV1)\n");
																		 break;

			default                      : printf("The Tag found on the reader is not supported.\n");
																		 if (p_batch && p_continue)
																		   goto batch_end;
			                               goto failed;
		}

    /* Writable, empty, ? */
		if (nfc_card_is_writable(card))
		{
			if (!p_overwrite_tag && !nfc_card_is_empty(card))
			{
				printf("The Tag already holds data. Try flag --overwrite-tag.");
			  if (p_batch && p_continue)
			    goto batch_end;
				goto failed;
			}
		} else
		{
      printf("The Tag is not writable (read only).\n");
			if (p_batch && p_continue)
			  goto batch_end;
      goto failed;
		}

		/* Capacity ? */
		if (ndef_tlv_size > nfc_card_get_capacity(card))
		{
      printf("This Tag is too small for the NDEF data.\n");
			if (p_batch && p_continue)
			  goto batch_end;
      goto failed;
		}

    /* Check pre-personalisation status */
		if (!nfc_card_is_prepared(card))
		{		
		  if (!p_prepare)
			{
        printf("The Tag is not prepared to store a NFC Tag NDEF. Try flag --prepare.\n");
			  if (p_batch && p_continue)
			    goto batch_end;
		    goto failed;
		  }

      /* Pre-personalize the card */
			if (!nfc_card_prepare(card))
			{
				printf("Failed to prepare the Tag.\n");
				if (p_batch && p_continue)
					goto batch_end;
				goto failed;
			}
    }


    /* Do write the card */
    if (!nfc_card_write_ndef(card, ndef_buffer, ndef_size))
		{
      printf("Failed to write the NDEF to the Tag.\n");
			if (p_batch && p_continue)
			  goto batch_end;
      goto failed;
		}

		/* Do lock the card */
		if (p_lock && !nfc_card_lock(card))
		{
      printf("Failed to lock the newly created Tag.\n");
			if (p_batch && p_continue)
			  goto batch_end;
      goto failed;
		}

    /* Done */
    nfc_card_close(card);
		card = NULL;

    if (!nfc_silent)
		  printf("NFC Tag has been successfully created.\n");

batch_end:

    if (p_batch)
		{
		  if (!nfc_silent)
			  printf("Remove the NFC Tag from the reader...\n");
			if (!nfc_reader_wait_card_absent(p_reader))
			{
				printf("Failed to wait for card removal.\n");
				goto failed;
			}

			goto batch_begin;
		}

	}

  /* Success */
  nfc_tag_free(tag);
  return EXIT_SUCCESS;
  
  /* Error */
failed:
  nfc_free(ndef_buffer);
  nfc_card_close(card);
  nfc_tag_free(tag);
	nfc_free(p_input);
	nfc_free(p_output);
	return EXIT_FAILURE;
}



static void usage(void)
{
  printf("\n");
  printf("Options to define a SmartPoster:\n\n");
  printf(" --uri <URI> , -u <URI>\n");
	printf("\tThe URI to put in the SmartPoster. This is mandatory.\n\n");
	printf(" --title <[Lang:]Text> , -t <[Lang:]Text>\n");
	printf(" [--title-lang <Lang>] --title-text <Text>\n");
	printf("\tThe Title of the the SmartPoster.\n");
	printf("\tYou may set different Title Texts for different Langs.\n\n");
	printf(" --action <open/save/edit> , -a <open/save/edit>\n");
	printf("\tThe action to be performed by the NFC device.\n\n");
	printf(" --mime-type <MINE-Type> , -m <MIME-Type>\n");
  printf("\tThe MIME-Type of the target pointed by the URI.\n\n");
	printf(" --size <Size> , -z <Size>\n");
  printf("\tThe Size (in bytes) of the target pointed by the URI.\n\n");
  printf(" --icon-file <Icon file> [--icon-type <Icon type>]\n");
	printf("\tThe Icon to be put in the SmartPoster.\n");
	printf("\tYou may specify the MIME-Type of the Icon too using --icon-type.\n");
	printf("\tIf unspecified, the software tries to guess the type according to\n");
	printf("\tthe file's extension and looking into file 'mime.types'.\n\n");
  printf("PC/SC readers:\n\n");
  printf(" --list-readers , -l\n");
	printf("\tPrint the list of available readers, then exits\n\n");
	printf(" --reader <Reader> , -r <Reader>\n");
	printf("\tWork with the specified Reader. It could be either the\n");
	printf("\tcomplete name of the Reader, or its index in the list.\n\n");
	printf("Tag encoding options:\n\n");
	printf(" --prepare , -p\n");
	printf("\tFormat the Tag according to NFC Forum specifications if needed.\n");
	printf("\tThis may be NOT REVERSIBLE (OTP bits). Use with care!\n\n");
	printf(" --lock , -k\n");
	printf("\tMake the tag READ-ONLY after encoding.\n");
	printf("\tThis may be NOT REVERSIBLE (Lock bits). Use with care!\n\n");
	printf(" --overwrite-tag\n\n");
	printf("\tOverwrite the data pre-existing in the Tag (if some).\n");
	printf("\tAny data present on the Tag WILL BE LOST. Use with care!\n\n");
  printf("Files:\n\n");
  printf(" --input-file <File> , -i <File>\n");
	printf("\tRead an existing SmartPoster NDEF from File\n");
	printf("\t(if no extension is specified, '.ndef' extension is assumed)\n\n");
	printf(" --output-file <File> , -o <File>\n");
	printf("\tWrite the generated SmartPoster NDEF to File\n\n");
	printf("\t(if no extension is specified, '.ndef' extension is assumed)\n\n");
	printf(" --overwrite-file\n");
	printf("\tOverwrite the output File is it already exists.\n\n");
	printf("Batch mode:\n\n");
	printf(" --batch , -b\n");
	printf("\tEncode more than one Tag.\n\n");
	printf(" --continue , -c\n");
	printf("\tDo not exit, even if errors occur.\n\n");
	printf("Misc:\n\n");
	printf(" --verbose , -v\n");
	printf("\tGive details on what's going on.\n\n");
	printf(" --quiet , -q\n");
	printf("\tShut up unless fatal errors occur.\n\n");
	printf(" --help , -h\n");
	printf("\tDisplay this help screen, then exits.\n\n");
}

static void hello(void)
{
  printf("%s: Create NFC Forum compliant Tags (SmartPoster)\n", prog_name);
	printf("\n");
  printf("This software is copyright (c) 2000-2011 PRO ACTIVE SAS.\n");
	printf("All rights reserved.\n");
  printf("Visit www.springcard.com/solutions/nfc for information and updates\n");
  printf("\n");
  printf("THIS SOFTWARE IS PROVIDED \"AS IS\" WITH NO EXPRESSED OR IMPLIED WARRANTY.\n");
  printf("You're using it at your own risks.\n");
  printf("\n");
}

static void helpmsg(const char *str)
{
  if (!nfc_silent)
	{
	  if (str != NULL)
		{
      printf(str);
			printf("\n");
		}
    printf("Call %s -h for Help\n", prog_name);
	}
}
