/**h* NFCTools/nfc_ndef.c
 *
 * NAME
 *   nfc_ndef.c
 *
 * DESCRIPTION
 *   Manipulation of NFC NDEF parsing and assembling
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"
#include "nfc_tag_i.h"

BOOL get_ndef(_ndef_t *ndef, size_t *offset, unsigned char *ndef_buffer, size_t ndef_size)
{
  int i;
  size_t o = *offset;

  ndef->header = ndef_buffer[o++];
  if (o >= ndef_size) goto overflow;

  if (nfc_verbose)
  {
    printf("\tNDEF header = %02X\t", ndef->header);
    if (ndef->header & NDEF_HEADER_MESSAGE_BEGIN)     printf("MB "); else printf("   ");
    if (ndef->header & NDEF_HEADER_MESSAGE_END)       printf("ME "); else printf("   ");
    if (ndef->header & NDEF_HEADER_CHUNK_FLAG)        printf("CF "); else printf("   ");
    if (ndef->header & NDEF_HEADER_SHORT_RECORD)      printf("SR "); else printf("   ");
    if (ndef->header & NDEF_HEADER_ID_LENGTH_PRESENT) printf("IL "); else printf("   ");
    printf("TNF=%d\n", ndef->header & NDEF_HEADER_TNF_MASK);
  }

  if ((o == 1) && (!(ndef->header & NDEF_HEADER_MESSAGE_BEGIN)))
  {
    if (!nfc_silent)
      printf("\tFirst header should have flag MB\n");
  }

  ndef->type_length = ndef_buffer[o++];
  if (o >= ndef_size) goto overflow;

  if (nfc_verbose)
    printf("\tType Length = %d\n", ndef->type_length);

  if (ndef->header & NDEF_HEADER_SHORT_RECORD)
  {
    ndef->payload_length  = ndef_buffer[o++];
  } else
  {
    ndef->payload_length   = 0;
    for (i=0; i<4; i++)
    {
      ndef->payload_length <<= 8;
      ndef->payload_length  += ndef_buffer[o++];
      if (o >= ndef_size) goto overflow;
    }
  }

  if (nfc_verbose)
    printf("\tPayload Length = %lu\n", (unsigned long) ndef->payload_length);

  if (ndef->header & NDEF_HEADER_ID_LENGTH_PRESENT)
  {
    ndef->id_length = ndef_buffer[o++];
    if (o >= ndef_size) goto overflow;
    
    if (nfc_verbose)
      printf("\tID Length = %lu\n", (unsigned long) ndef->id_length);

  } else
  {
    ndef->id_length = 0;
  }

  if (ndef->type_length)
  {
    ndef->type = (char *) &ndef_buffer[o];
    o += ndef->type_length;
    if (o >= ndef_size) goto overflow;
  } else
  {
    ndef->type = NULL;
  }
  if (ndef->id_length)
  {
    ndef->id = (char *) &ndef_buffer[o];
    o += ndef->id_length;
    if (o >= ndef_size) goto overflow;
  } else
  {
    ndef->id = NULL;
  }
  if (ndef->payload_length)
  {
    ndef->payload = &ndef_buffer[o];
    o += ndef->payload_length;
    if (o > ndef_size) goto overflow;
  } else
  {
    ndef->payload = NULL;
  }

  if ((o >= ndef_size) && (!(ndef->header & NDEF_HEADER_MESSAGE_END)))
  {
    if (!nfc_silent)
      printf("\tLast header should have flag ME\n");
  }

  *offset = o;
  return TRUE;

overflow:
  if (!nfc_silent)
    printf("NDEF overflow (@%lu > %lu)\n", (unsigned long) o, (unsigned long) ndef_size);
  return FALSE;
}

BOOL ndef_parse_smartposter(nfc_tag_t tag, unsigned char *ndef_buffer, size_t ndef_size)
{
  size_t offset = 0;
  int i;

  if ((tag == NULL) || (ndef_buffer == NULL) || (ndef_size < 6))
    return FALSE;

  if (nfc_verbose)
    printf("ndef.parse_smartposter (size=%lu)\n", (unsigned long) ndef_size);

  while (offset < ndef_size)
  {
    _ndef_t ndef;

    if (nfc_verbose)
      printf("ndef.parse_smartposter@%lu\n", (unsigned long) offset);

    if (!get_ndef(&ndef, &offset, ndef_buffer, ndef_size))
      return FALSE;

    switch (ndef.header & NDEF_HEADER_TNF_MASK)
    {
      case NDEF_HEADER_TNF_NFC_RTD_WKN :
        /* Well-known type */
        /* --------------- */
        if ((ndef.type_length == 1) && !strncmp(ndef.type, "U", 1))
        {
          /* URI */
          char *uri;

          if (!uri_decode_alloc_a(&uri, ndef.payload, ndef.payload_length))
          {
            if (!nfc_silent)
              printf("\tURI decode failed\n");
            return FALSE;
          }

          if (!nfc_smartposter_set_uri(tag, uri))
          {
            if (!nfc_silent)
              printf("\tURI assignment failed\n");
            nfc_free(uri);
            return FALSE;
          }

          nfc_free(uri);
        } else
        if ((ndef.type_length == 1) && !strncmp(ndef.type, "T", 1))
        {
          /* Text */
          char *title, *lang;

          if (!text_decode_alloc_a(&title, &lang, ndef.payload, ndef.payload_length))
          {
            if (!nfc_silent)
              printf("\tTitle decode failed\n");
            return FALSE;
          }

          if (!nfc_smartposter_set_title(tag, title, lang))
          {
            if (!nfc_silent)
              printf("\tTitle assignment failed\n");
            nfc_free(title);
            return FALSE;
          }

          nfc_free(title);
        } else
        if ((ndef.type_length == 1) && !strncmp(ndef.type, "t", 1))
        {
          /* MIME Type */
          char *mime_type = nfc_strdup_blob(ndef.payload, ndef.payload_length);

          if (!nfc_smartposter_set_mime_type(tag, mime_type))
          {
            if (!nfc_silent)
              printf("\tMIME-Type assignment failed\n");
            nfc_free(mime_type);
            return FALSE;
          }

          nfc_free(mime_type);
        } else
        if ((ndef.type_length == 1) && !strncmp(ndef.type, "s", 1))
        {
          /* Size */
          if (ndef.payload_length == 4)
          {
            unsigned long size = 0;

            for (i=0; i<4; i++)
            {
              size <<= 4;
              size  += ndef.payload[i];
            }

            if (!nfc_smartposter_set_size(tag, size))
            {
              if (!nfc_silent)
                printf("\tSize assignment failed\n");

              return FALSE;
            }
          }
        } else
        if ((ndef.type_length == 3) && !strncmp(ndef.type, "act", 3))
        {
          /* Action */
          if (ndef.payload_length == 1)
          {
            if (!nfc_smartposter_set_action(tag, ndef.payload[0]))
            {
              if (!nfc_silent)
                printf("\tAction assignment failed\n");
              return FALSE;
            }
          }
        }
        break;

      case NDEF_HEADER_TNF_MEDIA_TYPE :
        /* Media */
        /* ----- */

        if (!nfc_smartposter_set_icon_blob(tag, ndef.payload, ndef.payload_length))
        {
          if (!nfc_silent)
            printf("\tIcon assignment failed\n");
          return FALSE;
        }

        if (ndef.type_length)
        {
          char *icon_mime_type = nfc_strdup_blob(ndef.payload, ndef.payload_length);

          if (!nfc_smartposter_set_icon_mime_type(tag, icon_mime_type))
          {
            if (!nfc_silent)
              printf("\tIcon MIME-Type assignment failed\n");
            nfc_free(icon_mime_type);
            return FALSE;
          }

          nfc_free(icon_mime_type);     
        }

        break;
    }
  }

  return TRUE;
}


BOOL ndef_parse_tag(nfc_tag_t tag, unsigned char *ndef_buffer, size_t ndef_size)
{
  size_t offset = 0;
  int c = 0;

  if ((tag == NULL) || (ndef_buffer == NULL) || (ndef_size < 6))
    return FALSE;

  if (nfc_verbose)
    printf("ndef.parse_tag (size=%lu)\n", (unsigned long) ndef_size);

  while (offset < ndef_size)
  {
    _ndef_t ndef;

    if (nfc_verbose)
      printf("ndef.parse_tag@%lu\n", (unsigned long) offset);

    if (!get_ndef(&ndef, &offset, ndef_buffer, ndef_size))
      return FALSE;

    c++;

    switch (ndef.header & NDEF_HEADER_TNF_MASK)
    {
      case NDEF_HEADER_TNF_NFC_RTD_WKN : /* Well-known type */
        if ((ndef.type_length == 2) && (ndef.type[0] == 'S') && (ndef.type[1] == 'p'))
        {
          /* SmartPoster ;-) */
          if (!nfc_smartposter_init(tag))
            return FALSE;

          return ndef_parse_smartposter(tag, ndef.payload, ndef.payload_length);
        }
        break;

      default:
        if (!nfc_silent)
          printf("Unsupported NDEF record found\n");
    }
  }

  if (!c)
    if (!nfc_silent)
      printf("No NDEF record found\n");

  return FALSE;
}

BOOL ndef_assemble_smartposter(nfc_tag_t tag, unsigned char *ndef_buffer, size_t *ndef_size)
{
  _ndef_t *ndef_tree;
	_nfc_title_t *title;
  int i, count;
  size_t s;
  BOOL result = FALSE;

  if ((tag == NULL) || (tag->type != NFC_TAG_SMARTPOSTER))
    return FALSE;

  count = 1;
  if (tag->content.smartposter.uri != NULL)
    count++;
	title = tag->content.smartposter.titles;
	while (title != NULL)
	{
	  title = title->next;
    count++;
	}
  if (tag->content.smartposter.action >= 0)
    count++;
  if (tag->content.smartposter.mime_type != NULL)
    count++;
  if (tag->content.smartposter.size > 0)
    count++;
  if (tag->content.smartposter.icon_blob != NULL)
    count++;

  ndef_tree = nfc_alloc(count * sizeof(_ndef_t));

  ndef_tree[0].header = NDEF_HEADER_MESSAGE_BEGIN|NDEF_HEADER_MESSAGE_END|NDEF_HEADER_TNF_NFC_RTD_WKN;
  ndef_tree[0].type = nfc_strdup_length("Sp", &ndef_tree[0].type_length);

  count = 1;

  if (tag->content.smartposter.uri != NULL)
  {
    ndef_tree[count].header = NDEF_HEADER_TNF_NFC_RTD_WKN;    
    ndef_tree[count].type = nfc_strdup_length("U", &ndef_tree[count].type_length);

    if (!uri_encode_alloc_a(&ndef_tree[count].payload, &ndef_tree[count].payload_length, tag->content.smartposter.uri))
      goto cleanup;

    count++;
  }

	title = tag->content.smartposter.titles;
	while (title != NULL)
	{
    ndef_tree[count].header = NDEF_HEADER_TNF_NFC_RTD_WKN;
    ndef_tree[count].type = nfc_strdup_length("T", &ndef_tree[count].type_length);

    if (!text_encode_alloc_a(&ndef_tree[count].payload, &ndef_tree[count].payload_length, title->text, title->lang))
      goto cleanup;

	  title = title->next;
    count++;
	}

  if (tag->content.smartposter.action >= 0)
  {
    ndef_tree[count].header = NDEF_HEADER_TNF_NFC_RTD_WKN;
    ndef_tree[count].type = nfc_strdup_length("act", &ndef_tree[count].type_length);

    ndef_tree[count].payload_length = 1;
    ndef_tree[count].payload = nfc_alloc(ndef_tree[count].payload_length);

    ndef_tree[count].payload[0] = tag->content.smartposter.action;

    count++;
  }

  if (tag->content.smartposter.size > 0)
  {
    size_t v = tag->content.smartposter.size;

    ndef_tree[count].header = NDEF_HEADER_TNF_NFC_RTD_WKN;
    ndef_tree[count].type = nfc_strdup_length("s", &ndef_tree[count].type_length);

    ndef_tree[count].payload_length = 4;
    ndef_tree[count].payload = nfc_alloc(ndef_tree[count].payload_length);

    ndef_tree[count].payload[3] = (unsigned char) v; v >>= 8;
    ndef_tree[count].payload[2] = (unsigned char) v; v >>= 8;
    ndef_tree[count].payload[1] = (unsigned char) v; v >>= 8;
    ndef_tree[count].payload[0] = (unsigned char) v;

    count++;
  }

  if (tag->content.smartposter.mime_type != NULL)
  {
    ndef_tree[count].header = NDEF_HEADER_TNF_NFC_RTD_WKN;
    ndef_tree[count].type = nfc_strdup_length("t", &ndef_tree[count].type_length);
    ndef_tree[count].payload = (unsigned char *) nfc_strdup_length(tag->content.smartposter.mime_type, &ndef_tree[count].payload_length);

    count++;
  }

  if (tag->content.smartposter.icon_blob != NULL)
  {
    ndef_tree[count].header = NDEF_HEADER_TNF_MEDIA_TYPE;
    ndef_tree[count].type = nfc_strdup_length(tag->content.smartposter.icon_mime_type, &ndef_tree[count].type_length);

    ndef_tree[count].payload_length = tag->content.smartposter.icon_size;
    ndef_tree[count].payload = nfc_alloc(ndef_tree[count].payload_length);

    memcpy(ndef_tree[count].payload, tag->content.smartposter.icon_blob, tag->content.smartposter.icon_size);

    count++;
  }

  ndef_tree[1].header       |= NDEF_HEADER_MESSAGE_BEGIN;
  ndef_tree[count-1].header |= NDEF_HEADER_MESSAGE_END;

  for (i=1; i<count; i++)
  {
    ndef_tree[0].payload_length += 2; /* Header, Type Length */
    ndef_tree[0].payload_length += ndef_tree[i].type_length;

    if (ndef_tree[i].payload_length < 256)
    {
      ndef_tree[i].header |= NDEF_HEADER_SHORT_RECORD;
      ndef_tree[0].payload_length += 1; /* Length on 1B */
    } else
    {
      ndef_tree[0].payload_length += 4; /* Length on 4B */
    }
    ndef_tree[0].payload_length += ndef_tree[i].payload_length;

    if (ndef_tree[i].id_length)
    {
      ndef_tree[i].header |= NDEF_HEADER_ID_LENGTH_PRESENT;
      ndef_tree[0].payload_length += 1; /* ID Length on 1B */
      ndef_tree[0].payload_length += ndef_tree[i].id_length;
    }
  }

  s  = 2;
  s += ndef_tree[0].type_length;

  if (ndef_tree[0].payload_length < 256)
  {
    ndef_tree[0].header |= NDEF_HEADER_SHORT_RECORD;
    s += 1;
  } else
  {
    s += 4;
  }
  s += ndef_tree[0].payload_length;
  if (ndef_tree[0].id_length)
  {
    s += 1;
    s += ndef_tree[0].id_length;
  }

  if (ndef_size != NULL)
  {
    if ((ndef_buffer != NULL) && (*ndef_size < s))
      goto cleanup;

    *ndef_size = s;
  }

  if (ndef_buffer != NULL)
  {
    s = 0;
    ndef_buffer[s++] = ndef_tree[0].header;
    ndef_buffer[s++] = (unsigned char) ndef_tree[0].type_length;
    if (ndef_tree[0].header & NDEF_HEADER_SHORT_RECORD)
    {
      ndef_buffer[s++] = (unsigned char) ndef_tree[0].payload_length;
    } else
    {
      size_t v = ndef_tree[0].payload_length;
      ndef_buffer[s + 3] = (unsigned char) v; v >>= 8;
      ndef_buffer[s + 2] = (unsigned char) v; v >>= 8;
      ndef_buffer[s + 1] = (unsigned char) v; v >>= 8;
      ndef_buffer[s + 0] = (unsigned char) v;
      s += 4;
    }
    if (ndef_tree[0].header & NDEF_HEADER_ID_LENGTH_PRESENT)
    {
      ndef_buffer[s++] = (unsigned char) ndef_tree[0].id_length;
    }

    memcpy(&ndef_buffer[s], ndef_tree[0].type, ndef_tree[0].type_length);
    s += ndef_tree[0].type_length;

    if (ndef_tree[0].header & NDEF_HEADER_ID_LENGTH_PRESENT)
    {
      memcpy(&ndef_buffer[s], ndef_tree[0].id, ndef_tree[0].id_length);
      s += ndef_tree[0].id_length;
    }

    for (i=1; i<count; i++)
    {
      ndef_buffer[s++] = ndef_tree[i].header;
      ndef_buffer[s++] = (unsigned char) ndef_tree[i].type_length;
      if (ndef_tree[i].header & NDEF_HEADER_SHORT_RECORD)
      {
        ndef_buffer[s++] = (unsigned char) ndef_tree[i].payload_length;
      } else
      {
        size_t v = ndef_tree[i].payload_length;
        ndef_buffer[s + 3] = (unsigned char) v; v >>= 8;
        ndef_buffer[s + 2] = (unsigned char) v; v >>= 8;
        ndef_buffer[s + 1] = (unsigned char) v; v >>= 8;
        ndef_buffer[s + 0] = (unsigned char) v;
        s += 4;
      }
      if (ndef_tree[i].header & NDEF_HEADER_ID_LENGTH_PRESENT)
      {
        ndef_buffer[s++] = (unsigned char) ndef_tree[i].id_length;
      }

      memcpy(&ndef_buffer[s], ndef_tree[i].type, ndef_tree[i].type_length);
      s += ndef_tree[i].type_length;

      if (ndef_tree[i].header & NDEF_HEADER_ID_LENGTH_PRESENT)
      {
        memcpy(&ndef_buffer[s], ndef_tree[i].id, ndef_tree[i].id_length);
        s += ndef_tree[i].id_length;
      }

      memcpy(&ndef_buffer[s], ndef_tree[i].payload, ndef_tree[i].payload_length);
      s += ndef_tree[i].payload_length;
    }
  }

  result = TRUE;

cleanup:
  if (ndef_tree != NULL)
  {
    for (i=0; i<count; i++)
    {
      nfc_free(ndef_tree[i].id);
      nfc_free(ndef_tree[i].type);
      nfc_free(ndef_tree[i].payload);
    }
    nfc_free(ndef_tree);
  }
  return result;
}

BOOL ndef_assemble_tag_alloc(nfc_tag_t tag, unsigned char **ndef_buffer, size_t *ndef_size)
{
  size_t sz;

  if (tag == NULL)
    return FALSE;

  if (ndef_buffer != NULL)
    *ndef_buffer = NULL;
  if (ndef_size != NULL)
    *ndef_size = 0;

  if (tag->type == NFC_TAG_SMARTPOSTER)
  {
    if (!ndef_assemble_smartposter(tag, NULL, &sz))
    {
      if (!nfc_silent)
        printf("Failed to compute size of SmartPoster\n");
      return FALSE;
    }

    if (nfc_verbose)
      printf("SmartPoster size=%lu\n", (unsigned long) sz);

    if (ndef_size != NULL)
      *ndef_size = sz;

    if ((sz) && (ndef_buffer != NULL))
    {
      *ndef_buffer = nfc_alloc(sz);

      if (!ndef_assemble_smartposter(tag, *ndef_buffer, &sz))
      {
        if (!nfc_silent)
          printf("Failed to get the SmartPoster buffer\n");
        nfc_free(*ndef_buffer);
        *ndef_buffer = NULL;
        return FALSE;
      }
    }

    return TRUE;
  }

  return FALSE;
}
