/**h* NFCTools/nfc_type_2.c
 *
 * NAME
 *   nfc_type_1.c
 *
 * DESCRIPTION
 *   Support of NFC Type 2 Tag
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"
#include "nfc_card_i.h"

BOOL nfc_card_ultralight_recognize(nfc_card_t card)
{
	unsigned char buffer[16];
	unsigned char buffer_t[16];
	int i, page, pages;

  if (nfc_verbose)
		printf("\tthis is a chip in the Mifare UltraLight family\n");

	card->type = CARD_NXP_MIFARE_UL;
	card->capacity = 48;
	pages = 16;

  /* Recognize the chip */
  if (!nfc_card_read_memory(card,  0, buffer, 16))
	{
		if (nfc_verbose)
			printf("\tfailed to read pages 0-3\n");
		return FALSE;
	}
  if (!nfc_card_read_memory(card, 16, buffer_t, 16))
	{
		if (nfc_verbose)
			printf("\tpages 16-19 not readable, this must be a Mifare UltraLight\n");
	} else
	if (!memcmp(buffer, buffer_t, 16))
	{
		if (nfc_verbose)
			printf("\tpages 0-3 and 16-19 are equals, this must be a Mifare UltraLight\n");
	} else
	{
		if (nfc_verbose)
			printf("\tpages 0-3 and 16-19 are different, this must be a Mifare UltraLight C\n");
		card->type = CARD_NXP_MIFARE_UL_C;
		card->capacity = 144;
		pages = 40;
	}

	card->writable = TRUE;

	/* Analyze the CC - it must either be empty, or a valid CC */
	if ((buffer[12] == NFC_FORUM_MAGIC_NUMBER) && (buffer[13] != 0x00) && (buffer[14] != 0x00))
	{
		card->prepared = TRUE;
    card->capacity = 8 * buffer[14];

		if (nfc_verbose)
			printf("\tCC are valid, capacity=%dB\n", card->capacity);

		if ((buffer[15] & 0x0F) == 0x0F)
		{
			card->writable = FALSE;
			if (nfc_verbose)
        printf("\tCC says the card is not writable\n");
		}
	} else
	if ((buffer[12] == 0x00) || (buffer[13] == 0x00) || (buffer[14] == 0x00) || (buffer[15] == 0x00))
	{
		if (nfc_verbose)
      printf("\tCC are blank\n");
	} else
	{
		card->writable = FALSE;
		if (nfc_verbose)
      printf("\tCC are invalid\n");
	}

	/* Check the locks */
	if (card->writable)
	{
		if (buffer[10] || buffer[11])
		{
		  card->writable = FALSE;
			if (nfc_verbose)
				printf("\tlock bits of page 3 are set\n");
		}
	}
	if (card->writable && (card->type = CARD_NXP_MIFARE_UL_C))
	{
		if (!nfc_card_read_memory(card, 40, buffer, 16))
		{
			if (nfc_verbose)
				printf("\tread page 40 failed\n");
		} else
		if (buffer[0] || buffer[1])
		{
		  card->writable = FALSE;
			if (nfc_verbose)
				printf("\tlock bits of page 40 are set\n");
		}
	}

	/* Is card empty ? */
	card->empty = TRUE;
	for (page=4; page<pages; page+=4)
	{
		if (!nfc_card_read_memory(card, (unsigned short) page, buffer, 16))
		{
			if (nfc_verbose)
				printf("\tread page 40 failed\n");
			break;
		}

		for (i=0; i<16; i++)
		{
			if (buffer[i] != 0x00)
			{
				if (nfc_verbose)
				  printf("\tcard is not empty\n");
				card->empty = FALSE;
				break;
      }
		}

		printf("\n");

		if (!card->empty)
			break;
	}

	return TRUE;
}

BOOL nfc_card_ultralight_prepare(nfc_card_t card)
{
  BYTE buffer[4];

  if ((card == NULL) || ((card->type != CARD_NXP_MIFARE_UL) && (card->type != CARD_NXP_MIFARE_UL_C)))
    return FALSE;

	buffer[0] = NFC_FORUM_MAGIC_NUMBER;
	buffer[1] = NFC_FORUM_VERSION_NUMBER;
	buffer[2] = card->capacity / 8;
	buffer[3] = 0x00;

	if (nfc_verbose)
		printf("\twriting the CC on page 3\n");

	if (!nfc_card_write_memory(card, 3, buffer, 4))
	{
	  if (!nfc_silent)
		  printf("Failed to write the CC into the UltraLight%s card\n", (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
		return FALSE;
	}

	return TRUE;
}

BOOL nfc_card_ultralight_lock(nfc_card_t card)
{
  BYTE buffer[4];

  if ((card == NULL) || ((card->type != CARD_NXP_MIFARE_UL) && (card->type != CARD_NXP_MIFARE_UL_C)))
    return FALSE;

  if (card->type == CARD_NXP_MIFARE_UL_C)
	{
    buffer[0] = 0xFF;
		buffer[1] = 0xFF;
		buffer[2] = 0x00;
		buffer[3] = 0x00;

		if (!nfc_card_write_memory(card, 40, buffer, 4))
		{
			if (!nfc_silent)
				printf("Failed to write page 40 of the UltraLight C card\n");
			return FALSE;
		}
	}

  buffer[0] = 0x00;
	buffer[1] = 0x00;
	buffer[2] = 0xFF;
	buffer[3] = 0xFF;

	if (!nfc_card_write_memory(card, 2, buffer, 4))
	{
		if (!nfc_silent)
			printf("Failed to write page 2 of the UltraLight%s card\n", (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
		return FALSE;
	}

  buffer[0] = 0x00;
	buffer[1] = 0x00;
	buffer[2] = 0x00;
	buffer[3] = 0x0F;

	if (!nfc_card_write_memory(card, 3, buffer, 4))
	{
		if (!nfc_silent)
			printf("Failed to write page 3 of the UltraLight%s card\n", (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
		return FALSE;
	}

  return TRUE;
}

BOOL nfc_card_ultralight_write_content(nfc_card_t card, unsigned char *content, size_t content_length)
{
  BYTE buffer[4];
	int page, last_page;
	size_t offset = 0;

  if ((card == NULL) || ((card->type != CARD_NXP_MIFARE_UL) && (card->type != CARD_NXP_MIFARE_UL_C)))
    return FALSE;

  last_page = 4 + card->capacity / 4;
	for (page=4; page<last_page; page++)
	{
	  if (offset > content_length)
		  break;

	  memcpy(buffer, &content[offset], 4);
    offset += 4;

		if (!nfc_card_write_memory(card, (unsigned short) page, buffer, 4))
		{
			if (!nfc_silent)
				printf("Failed to write page %d of the UltraLight%s card\n", page, (card->type == CARD_NXP_MIFARE_UL_C) ? " C" : "");
			return FALSE;
		}
	}

  return TRUE;
}

static const char *sorry = "Sorry, handling of Mifare Classic cards is not yet implemented\n";

BOOL nfc_card_mifare_recognize(nfc_card_t card)
{
  (void) (card);

  printf(sorry);
  return FALSE;
}

BOOL nfc_card_mifare_prepare(nfc_card_t card)
{
  (void) (card);

  printf(sorry);
  return FALSE;
}

BOOL nfc_card_mifare_lock(nfc_card_t card)
{
  (void) (card);

  printf(sorry);
  return FALSE;
}

BOOL nfc_card_mifare_write_content(nfc_card_t card, unsigned char *content, size_t content_length)
{
  (void) (card);
	(void) (content);
	(void) (content_length);

  printf(sorry);
  return FALSE;
}
