/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2003-2011 PRO ACTIVE SAS - www.springcard.com
   
  ref_mifare_pcsc.c
  -----------------

*/
#include "springprox_pcsc.h"

typedef enum
{
  RC_TRUE,
  RC_FALSE,

  RC_PARAMETER_ERROR,
  RC_ACCEPTABLE_ERROR,
  RC_FATAL_ERROR,

} RC_T;


/* Interpreted APDUs defined by the PC/SC standard */
/* ----------------------------------------------- */

RC_T SCardLoadKey(SCARDHANDLE hCard, BYTE key_location, BYTE key_index, const BYTE key_value[], BYTE key_size);
RC_T SCardGeneralAuthenticate(SCARDHANDLE hCard, BYTE address, BYTE key_location, BYTE key_index);
RC_T SCardReadBinary(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size);
RC_T SCardUpdateBinary(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size);

/* Interpreted APDUs out of the PC/SC standard */
/* ------------------------------------------- */

RC_T SCardMifareClassicRead_Auto(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size);
RC_T SCardMifareClassicRead_KeyIndex(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size, BYTE key_location, BYTE key_index);
RC_T SCardMifareClassicRead_KeyValue(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size, const BYTE key_value[6]);

RC_T SCardMifareClassicWrite_Auto(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size);
RC_T SCardMifareClassicWrite_KeyIndex(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size, BYTE key_location, BYTE key_index);
RC_T SCardMifareClassicWrite_KeyValue(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size, const BYTE key_value[6]);

/* Mifare helper */
/* ------------- */

void MakeMifareSectorTrailer(BYTE trailer[16], const BYTE key_a[6], const BYTE key_b[6], const BYTE ac[4]);

/* Test helpers */
/* ------------ */

RC_T SCardLoadKeys(SCARDHANDLE hCard);
RC_T SCardMifareIsTransport(SCARDHANDLE hCard);
RC_T SCardMifareFormatTransport(SCARDHANDLE hCard);
RC_T SCardMifareFormatUsage(SCARDHANDLE hCard);

/* Test core */
/* --------- */

BOOL MifareTestCore(SCARDCONTEXT hContext, const char *szReader, const BYTE atr[], DWORD atrlen);

/* Keyring */
/* ------- */

extern const BYTE KeyTransport[6];
extern const BYTE KeyTestA[6];
extern const BYTE KeyTestB[6];

/* Constants */
/* --------- */

#define APDU_BUFFERS_SZ 280

#define KeyLocation    0x00  /* We put the keys in volatile memory */
#define KeyTestAIdx    0x00  /* Index for test key A (type A, offset 0) */
#define KeyTestBIdx    0x10  /* Index for test key B (type B, offset 0) */
#define KeyTestTIdx    0x01  /* Index for transport key (type A, offset 1) */

#ifdef WIN32
void StartInterval(void);
DWORD GetInterval(void);
#endif







int main(int argc, char **argv)
{
	SCARDCONTEXT      hContext;
  BOOL              bContextOK = FALSE;
	LPSTR             szReaders = NULL;
  DWORD             dwReadersSz;
	LPTSTR            pReader;
  SCARD_READERSTATE rgscState;

	LONG   rc;

	UNUSED_PARAMETER(argc);
	UNUSED_PARAMETER(argv);

  /*
   * Welcome message, check parameters
   * ---------------------------------
   */
  printf("\n");
  printf("SpringCard PC/SC SDK : Mifare Benchmark\n");
  printf("---------------------------------------\n");
  printf("v.090226 (c) 2008 Pro Active\n");
	printf("Press <Ctrl+C> to exit\n\n");

	/*
   * Get a handle to the PC/SC resource manager
   */
	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	if(rc != SCARD_S_SUCCESS)
	{
		printf("SCardEstablishContext error %08lX\n", rc);
		goto leave;
	}
  bContextOK = TRUE;

  /*
   * Get the list of available readers
   */
  dwReadersSz = SCARD_AUTOALLOCATE;
	rc = SCardListReaders(hContext,
                        NULL,                /* Any group */
                        (LPTSTR) &szReaders, /* Diabolic cast for buffer auto-allocation */
                        &dwReadersSz);       /* Beg for auto-allocation */
	  
	if (rc == SCARD_E_NO_READERS_AVAILABLE)
	{
		/* No reader at all */
		printf("No PC/SC reader\n");
    goto leave;
	}
  
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardListReaders error %08lX\n",rc);
		goto leave;
	}

  /*
   * Loop withing reader(s) to find a Mifare card
   */ 
  pReader = szReaders;
	while (*pReader != '\0')
	{
    /* Got a reader name */
    printf("\nPC/SC reader '%s' :\n", pReader);

    /* Get status */
    rgscState.szReader = pReader;
    rgscState.dwCurrentState = SCARD_STATE_UNAWARE;
 		rc = SCardGetStatusChange(hContext,
			                        0,
			                        &rgscState,
			                        1);	

    if (rc != SCARD_S_SUCCESS)
	  {
      printf("SCardGetStatusChange error %08lX\n", rc);
		  goto leave;
	  }

    /* Show reader's status */
    if (rgscState.dwEventState & SCARD_STATE_IGNORE)
      printf("\tIgnore this reader\n");

    if (rgscState.dwEventState & SCARD_STATE_UNKNOWN)
      printf("\tReader unknown\n");

    if (rgscState.dwEventState & SCARD_STATE_UNAVAILABLE)
      printf("\tStatus unavailable\n");

    if (rgscState.dwEventState & SCARD_STATE_EMPTY)
      printf("\tNo card in the reader\n");

    if (rgscState.dwEventState & SCARD_STATE_PRESENT)
      printf("\tCard present\n");

    /* Show ATR (if some) */
    if (rgscState.cbAtr)
		{
      BYTE i;

			printf("\tCard ATR: ");
			for (i=0; i<rgscState.cbAtr; i++)
				printf("%02X", rgscState.rgbAtr[i]);
      printf("\n");
    }

    /* Show card's status */
    if (rgscState.dwEventState & SCARD_STATE_ATRMATCH)
      printf("\tATR match\n");

    if (rgscState.dwEventState & SCARD_STATE_INUSE)
      printf("\tCard (or reader) in use\n");

    if (rgscState.dwEventState & SCARD_STATE_EXCLUSIVE)
      printf("\tCard (or reader) reserved for exclusive use\n");

    if (rgscState.dwEventState & SCARD_STATE_MUTE)
      printf("\tCard is mute\n");

    if ( (rgscState.dwEventState & SCARD_STATE_PRESENT)
     && !(rgscState.dwEventState & SCARD_STATE_MUTE)
     && !(rgscState.dwEventState & SCARD_STATE_INUSE)
     && !(rgscState.dwEventState & SCARD_STATE_EXCLUSIVE))
    {
      /* Call test core function */
      MifareTestCore(hContext, pReader, rgscState.rgbAtr, rgscState.cbAtr);
    }

    /* Jump to next entry in multi-string array */
		pReader += strlen(pReader) + 1;
	}

  printf("Exiting...\n");

leave:
  /* Free the list of readers  */
  if (szReaders != NULL)
  {
    SCardFreeMemory(hContext, szReaders);
    szReaders = NULL;
  }

  if (bContextOK)
  {      
	  SCardReleaseContext(hContext);
  }

  return EXIT_SUCCESS;
}



BOOL MifareTest_Standard_Blocks(SCARDHANDLE hCard, BOOL transport);
BOOL MifareTest_Standard_Sectors(SCARDHANDLE hCard, BOOL transport);

BOOL MifareTest_Specific_Blocks_A(SCARDHANDLE hCard, BOOL transport);
BOOL MifareTest_Specific_Sectors_A(SCARDHANDLE hCard, BOOL transport);

BOOL MifareTest_Specific_Blocks_K(SCARDHANDLE hCard, BOOL transport);
BOOL MifareTest_Specific_Sectors_K(SCARDHANDLE hCard, BOOL transport);

BOOL MifareTest_Specific_Blocks_I(SCARDHANDLE hCard, BOOL transport);
BOOL MifareTest_Specific_Sectors_I(SCARDHANDLE hCard, BOOL transport);

BOOL MifareTestCore(SCARDCONTEXT hContext, const char *szReader, const BYTE atr[], DWORD atrlen)
{
  SCARDHANDLE hCard;
  DWORD dwProtocol;
  LONG rc;

	UNUSED_PARAMETER(atr);
	UNUSED_PARAMETER(atrlen);

  /* Connect to the card */
  rc = SCardConnect(hContext, szReader, SCARD_SHARE_EXCLUSIVE, SCARD_PROTOCOL_T1, &hCard, &dwProtocol);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardConnect error %08lX\n", rc);
		return FALSE;
	}

  /* Load the keys we'll need into reader's memory */
  rc = SCardLoadKeys(hCard);
  if (rc != RC_TRUE)
  {
    printf("Failed to load the keys\n");
    goto failed;
  }

  /* Determine whether the card is already in transport conditions, or not */
  printf("Reading the card, please wait...\n");
  rc = SCardMifareIsTransport(hCard);
  if (rc == RC_FALSE)
  {
    printf("This card is not in transport condition !\n\n");

    if (0)
    {
      printf("Test is interrupted to protect the data that are currently on the card.\n");
      printf("Please run this tool again, with parameter -f to format the card.\n");
      goto failed;
    }
  } else
  if (rc != RC_TRUE)
  {
    printf("Failed to verify whether the card is in transport conditions, or not\n");
    goto failed;
  }

  /* Format the card to make it ready for the first tests */
#if 1
  printf("Formating the card, please wait...\n");
  rc = SCardMifareFormatUsage(hCard);
  if (rc != RC_TRUE)
  {
    printf("Failed to format the card\n");
    goto failed;
  }
#endif

  /* OK, let's run the first serie of tests */
#if 1
  if (!MifareTest_Standard_Blocks(hCard, FALSE))
    goto failed;
  if (!MifareTest_Standard_Sectors(hCard, FALSE))
    goto failed;
#endif

#if 1
  if (!MifareTest_Specific_Blocks_A(hCard, FALSE))
    goto failed;
  if (!MifareTest_Specific_Sectors_A(hCard, FALSE))
    goto failed;
#endif

#if 1
  if (!MifareTest_Specific_Blocks_K(hCard, FALSE))
    goto failed;
  if (!MifareTest_Specific_Sectors_K(hCard, FALSE))
    goto failed;
#endif

#if 1
  if (!MifareTest_Specific_Blocks_I(hCard, FALSE))
    goto failed;
  if (!MifareTest_Specific_Sectors_I(hCard, FALSE))
    goto failed;
#endif

#if 1
  /* Put back the card into transport condition */
  printf("Formating the card, please wait...\n");
  rc = SCardMifareFormatTransport(hCard);
  if (rc != RC_TRUE)
  {
    printf("Failed to put back the card in transport conditions\n");
    goto failed;
  }
#endif

  /* OK, let's run the second serie of tests */
#if 1
  if (!MifareTest_Standard_Blocks(hCard, TRUE))
    goto failed;
  if (!MifareTest_Standard_Sectors(hCard, TRUE))
    goto failed;
#endif

#if 1
  if (!MifareTest_Specific_Blocks_A(hCard, TRUE))
    goto failed;
  if (!MifareTest_Specific_Sectors_A(hCard, TRUE))
    goto failed;
#endif

#if 1
  if (!MifareTest_Specific_Blocks_K(hCard, TRUE))
    goto failed;
  if (!MifareTest_Specific_Sectors_K(hCard, TRUE))
    goto failed;
#endif

#if 1
  if (!MifareTest_Specific_Blocks_I(hCard, TRUE))
    goto failed;
  if (!MifareTest_Specific_Sectors_I(hCard, TRUE))
    goto failed;
#endif

  /* OK, done */
  printf("Done!\n");
  SCardDisconnect(hCard, SCARD_RESET_CARD);
  return TRUE;

failed:
  SCardDisconnect(hCard, SCARD_RESET_CARD);
  return FALSE;
}


/*
 * MifareTest_Standard_Blocks
 * --------------------------
 * Benchmark using standard PC/SC commands (GENERAL_AUTHENTICATE + READ_BINARY
 * + UPDATE_BINARY) at block level.
 */
BOOL MifareTest_Standard_Blocks(SCARDHANDLE hCard, BOOL transport)
{
  BYTE block_data[16];
  BYTE sector, block;
  RC_T rc;

  BYTE key_index = transport ? KeyTestTIdx : KeyTestBIdx;

  if (transport)
    printf("Test : STANDARD,BLOCKS - transport conditions\n");
  else
    printf("Test : STANDARD,BLOCKS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    /* Get authenticated on the sector */
    rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, key_index);
    if (rc != RC_TRUE)
    {
      printf("Authentication on sector %d failed\n", sector);
      return FALSE;
    }

    for (block = 0; block < 4; block++)
    {
      /* Read the block (even if it is a sector trailer) */
      rc = SCardReadBinary(hCard, (BYTE) ((4 * sector) + block), block_data, 16);
      if (rc != RC_TRUE)
      {
        printf("Read on sector %d, block %d failed\n", sector, block);
        return FALSE;
      }

      if ((block < 3) && ((sector != 0) || (block != 0)))
      {
        /* Write the block (appart sector trailers and read-only block 0) */
        rc = SCardUpdateBinary(hCard, (BYTE) ((4 * sector) + block), block_data, 16);
        if (rc != RC_TRUE)
        {
          printf("Write on sector %d, block %d failed\n", sector, block);
          return FALSE;
        }
      }
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}

/*
 * MifareTest_Standard_Sectors
 * ---------------------------
 * Benchmark using standard PC/SC commands (GENERAL_AUTHENTICATE + READ_BINARY
 * + UPDATE_BINARY) at sector level.
 */
BOOL MifareTest_Standard_Sectors(SCARDHANDLE hCard, BOOL transport)
{
  BYTE sector_data[64];
  BYTE sector;
  RC_T rc;
  
  BYTE key_index = transport ? KeyTestTIdx : KeyTestBIdx;
  
  if (transport)
    printf("Test : STANDARD,SECTORS - transport conditions\n");
  else
    printf("Test : STANDARD,SECTORS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    /* Get authenticated on the sector */
    rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, key_index);
    if (rc != RC_TRUE)
    {
      printf("Authentication on sector %d failed\n", sector);
      return FALSE;
    }

    /* Read the sector (data and masked trailer) */
    rc = SCardReadBinary(hCard, (BYTE) (4 * sector), sector_data, 64);
    if (rc != RC_TRUE)
    {
      printf("Read on sector %d failed\n", sector);
      return FALSE;
    }

    /* Write the sector (data only) */
    rc = SCardUpdateBinary(hCard, (BYTE) (4 * sector), sector_data, 48);
    if (rc != RC_TRUE)
    {
      printf("Write on sector %d failed\n", sector);
      return FALSE;
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}


/*
 * MifareTest_Specific_Blocks_A
 * ----------------------------
 * Benchmark using specific PC/SC commands (MIFARE_CLASSIC_READ
 * + MIFARE_CLASSIC_WRITE) at block level. The reader tries to
 * find the correct key automatically.
 */
BOOL MifareTest_Specific_Blocks_A(SCARDHANDLE hCard, BOOL transport)
{
  BYTE block_data[16];
  BYTE sector, block;
  RC_T rc;

  if (transport)
    printf("Test : SPECIFIC,AUTO,BLOCKS - transport conditions\n");
  else
    printf("Test : SPECIFIC,AUTO,BLOCKS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    for (block = 0; block < 4; block++)
    {
      /* Read the block (even if it is a sector trailer) */
      rc = SCardMifareClassicRead_Auto(hCard, (BYTE) ((4 * sector) + block), block_data, 16);
      if (rc != RC_TRUE)
      {
        printf("Read on sector %d, block %d failed\n", sector, block);
        return FALSE;
      }

      if ((block < 3) && ((sector != 0) || (block != 0)))
      {
        /* Write the block (appart sector trailers and read-only block 0) */
        rc = SCardMifareClassicWrite_Auto(hCard, (BYTE) ((4 * sector) + block), block_data, 16);
        if (rc != RC_TRUE)
        {
          printf("Write on sector %d, block %d failed\n", sector, block);
          return FALSE;
        }
      }
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}

/*
 * MifareTest_Specific_Sectors_A
 * -----------------------------
 * Benchmark using specific PC/SC commands (MIFARE_CLASSIC_READ
 * + MIFARE_CLASSIC_WRITE) at sector level. The reader tries to
 * find the correct key automatically.
 */
BOOL MifareTest_Specific_Sectors_A(SCARDHANDLE hCard, BOOL transport)
{
  BYTE sector_data[64];
  BYTE sector;
  RC_T rc;
   
  if (transport)
    printf("Test : SPECIFIC,AUTO,SECTORS - transport conditions\n");
  else
    printf("Test : SPECIFIC,AUTO,SECTORS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    /* Read the sector (data and masked trailer) */
    rc = SCardMifareClassicRead_Auto(hCard, (BYTE) (4 * sector), sector_data, 64);
    if (rc != RC_TRUE)
    {
      printf("Read on sector %d failed\n", sector);
      return FALSE;
    }

    /* Write the sector (data only) */
    rc = SCardMifareClassicWrite_Auto(hCard, (BYTE) (4 * sector), sector_data, 48);
    if (rc != RC_TRUE)
    {
      printf("Write on sector %d failed\n", sector);
      return FALSE;
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}


/*
 * MifareTest_Specific_Blocks_K
 * ----------------------------
 * Benchmark using specific PC/SC commands (MIFARE_CLASSIC_READ
 * + MIFARE_CLASSIC_WRITE) at block level. The value of the key
 * is passed on every function call.
 */
BOOL MifareTest_Specific_Blocks_K(SCARDHANDLE hCard, BOOL transport)
{
  BYTE block_data[16];
  BYTE sector, block;
  RC_T rc;

  const BYTE *key_value = transport ? KeyTransport : KeyTestB;

  if (transport)
    printf("Test : SPECIFIC,KEY VALUES,BLOCKS - transport conditions\n");
  else
    printf("Test : SPECIFIC,KEY VALUES,BLOCKS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    for (block = 0; block < 4; block++)
    {
      /* Read the block (even if it is a sector trailer) */
      rc = SCardMifareClassicRead_KeyValue(hCard, (BYTE) ((4 * sector) + block), block_data, 16, key_value);
      if (rc != RC_TRUE)
      {
        printf("Read on sector %d, block %d failed\n", sector, block);
        return FALSE;
      }

      if ((block < 3) && ((sector != 0) || (block != 0)))
      {
        /* Write the block (appart sector trailers and read-only block 0) */
        rc = SCardMifareClassicWrite_KeyValue(hCard, (BYTE) ((4 * sector) + block), block_data, 16, key_value);
        if (rc != RC_TRUE)
        {
          printf("Write on sector %d, block %d failed\n", sector, block);
          return FALSE;
        }
      }
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}

/*
 * MifareTest_Specific_Sectors_K
 * -----------------------------
 * Benchmark using specific PC/SC commands (MIFARE_CLASSIC_READ
 * + MIFARE_CLASSIC_WRITE) at sector level. The value of the key
 * is passed on every function call.
 */
BOOL MifareTest_Specific_Sectors_K(SCARDHANDLE hCard, BOOL transport)
{
  BYTE sector_data[64];
  BYTE sector;
  RC_T rc;
  
  const BYTE *key_value = transport ? KeyTransport : KeyTestB;
  
  if (transport)
    printf("Test : SPECIFIC,KEY VALUES,SECTORS - transport conditions\n");
  else
    printf("Test : SPECIFIC,KEY VALUES,SECTORS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    /* Read the sector (data and masked trailer) */
    rc = SCardMifareClassicRead_KeyValue(hCard, (BYTE) (4 * sector), sector_data, 64, key_value);
    if (rc != RC_TRUE)
    {
      printf("Read on sector %d failed\n", sector);
      return FALSE;
    }

    /* Write the sector (data only) */
    rc = SCardMifareClassicWrite_KeyValue(hCard, (BYTE) (4 * sector), sector_data, 48, key_value);
    if (rc != RC_TRUE)
    {
      printf("Write on sector %d failed\n", sector);
      return FALSE;
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}

/*
 * MifareTest_Specific_Blocks_I
 * ----------------------------
 * Benchmark using specific PC/SC commands (MIFARE_CLASSIC_READ
 * + MIFARE_CLASSIC_WRITE) at block level. The index of the key
 * is passed on every function call.
 */
BOOL MifareTest_Specific_Blocks_I(SCARDHANDLE hCard, BOOL transport)
{
  BYTE block_data[16];
  BYTE sector, block;
  RC_T rc;

  BYTE key_index = transport ? KeyTestTIdx : KeyTestBIdx;

  if (transport)
    printf("Test : SPECIFIC,KEY INDEX,BLOCKS - transport conditions\n");
  else
    printf("Test : SPECIFIC,KEY INDEX,BLOCKS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    for (block = 0; block < 4; block++)
    {
      /* Read the block (even if it is a sector trailer) */
      rc = SCardMifareClassicRead_KeyIndex(hCard, (BYTE) ((4 * sector) + block), block_data, 16, KeyLocation, key_index);
      if (rc != RC_TRUE)
      {
        if (rc == RC_ACCEPTABLE_ERROR)
        {
          printf("This version of reader's firmware doesn't support this new feature\n");
          return TRUE;
        }
        printf("Read on sector %d, block %d failed\n", sector, block);
        return FALSE;
      }

      if ((block < 3) && ((sector != 0) || (block != 0)))
      {
        /* Write the block (appart sector trailers and read-only block 0) */
        rc = SCardMifareClassicWrite_KeyIndex(hCard, (BYTE) ((4 * sector) + block), block_data, 16, KeyLocation, key_index);
        if (rc != RC_TRUE)
        {
          if (rc == RC_ACCEPTABLE_ERROR)
          {
            printf("This version of reader's firmware doesn't support this new feature\n");
            return TRUE;
          }
          printf("Write on sector %d, block %d failed\n", sector, block);
          return FALSE;
        }
      }
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}

/*
 * MifareTest_Specific_Sectors_K
 * -----------------------------
 * Benchmark using specific PC/SC commands (MIFARE_CLASSIC_READ
 * + MIFARE_CLASSIC_WRITE) at sector level. The index of the key
 * is passed on every function call.
 */
BOOL MifareTest_Specific_Sectors_I(SCARDHANDLE hCard, BOOL transport)
{
  BYTE sector_data[64];
  BYTE sector;
  RC_T rc;
  
  BYTE key_index = transport ? KeyTestTIdx : KeyTestBIdx;
  
  if (transport)
    printf("Test : SPECIFIC,KEY INDEX,SECTORS - transport conditions\n");
  else
    printf("Test : SPECIFIC,KEY INDEX,SECTORS - usage conditions\n");

#ifdef WIN32
  StartInterval();
#endif

  for (sector = 0; sector < 16; sector++)
  {
    /* Read the sector (data and masked trailer) */
    rc = SCardMifareClassicRead_KeyIndex(hCard, (BYTE) (4 * sector), sector_data, 64, KeyLocation, key_index);
    if (rc != RC_TRUE)
    {
      if (rc == RC_ACCEPTABLE_ERROR)
      {
        printf("This version of reader's firmware doesn't support this new feature\n");
        return TRUE;
      }
      printf("Read on sector %d failed\n", sector);
      return FALSE;
    }

    /* Write the sector (data only) */
    rc = SCardMifareClassicWrite_KeyIndex(hCard, (BYTE) (4 * sector), sector_data, 48, KeyLocation, key_index);
    if (rc != RC_TRUE)
    {
      if (rc == RC_ACCEPTABLE_ERROR)
      {
        printf("This version of reader's firmware doesn't support this new feature\n");
        return TRUE;
      }
      printf("Write on sector %d failed\n", sector);
      return FALSE;
    }
  }

#ifdef WIN32
  printf("Elapsed time : %ldms\n", GetInterval());
#endif
  return TRUE;
}




BOOL prefer_specific = FALSE;

/**f* PCSC/SCardLoadKey
 *
 * NAME
 *   SCardLoadKey
 *
 * DESCRIPTION
 *   Load a Mifare authentication key into the reader.
 *
 * INPUTS
 *   SCARDHANDLE hCard       : handle to the PC/SC smartcard
 *   BYTE key_location       : set to 0x00 to put the key in volatile memory
 *                             set to 0x20 to put the key in non-volatile memory
 *   BYTE key_index          : 0x00 to 0x0F are 'A' keys
 *                             0x10 to 0x1F are 'B' keys
 *   const BYTE key_value[]  : value of the key
 *   BYTE key_size           : must be 6 for a Mifare key
 *
 * OUTPUTS
 *   OK or error
 *
 * WARNING
 *   Non-volatile keys are stored in the E2PROM of the RC chipset.
 *   The E2PROM has a limited write endurance, therefore you mustn't overwrite
 *   the permanent keys "too often".
 *
 *   Typically, load your application keys in E2PROM once, when installating the
 *   product install, and then forget them...
 *
 * SEE ALSO
 *   SCardGeneralAuthenticate
 *
 **/
RC_T SCardLoadKey(SCARDHANDLE hCard, BYTE key_location, BYTE key_index, const BYTE key_value[], BYTE key_size)
{
  BYTE  pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (key_value == NULL) return RC_PARAMETER_ERROR;
  if (key_size + 5 > APDU_BUFFERS_SZ) return RC_PARAMETER_ERROR;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0x82;
  pbSendBuffer[2] = key_location;
  pbSendBuffer[3] = key_index;
  pbSendBuffer[4] = key_size;

  memcpy(&pbSendBuffer[5], key_value, key_size);

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, key_size+5, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(LOAD_KEY) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if (pcbRecvLength != 2)
  {
    printf("SCardTransmit(LOAD_KEY) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pbRecvBuffer[0] != 0x90) || (pbRecvBuffer[1] != 0x00))
  {
    printf("LOAD_KEY failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FATAL_ERROR;
  }

  return RC_TRUE;
}

/**f* PCSC/SCardGeneralAuthenticate
 *
 * NAME
 *   SCardGeneralAuthenticate
 *
 * DESCRIPTION
 *   Perform Mifare authentication, using one of the keys already loaded in the
 *   reader, on the specified block.
 *
 * INPUTS
 *   SCARDHANDLE hCard       : handle to the PC/SC smartcard
 *   BYTE address            : block number to get authenticated on
 *                             (remember: the __block__, not the __sector__)
 *   BYTE key_location       : as specified in SCardLoadKey
 *   BYTE key_index          : as specified in SCardLoadKey
 *
 * OUTPUTS
 *   OK or error
 *
 * SEE ALSO
 *   SCardLoadKey
 *
 **/
RC_T SCardGeneralAuthenticate(SCARDHANDLE hCard, BYTE address, BYTE key_location, BYTE key_index)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0x86;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = 0x00;
  pbSendBuffer[4] = 0x05;

  pbSendBuffer[5] = 0x01;
  pbSendBuffer[6] = 0x00;
  pbSendBuffer[7] = address;
  pbSendBuffer[8] = key_location;
  pbSendBuffer[9] = key_index;

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 10, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(GENERAL_AUTHENTICATE) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if (pcbRecvLength != 2)
  {
    printf("SCardTransmit(GENERAL_AUTHENTICATE) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pbRecvBuffer[0] != 0x90) || (pbRecvBuffer[1] != 0x00))
  {
    if ((pbRecvBuffer[0] != 0x69) || (pbRecvBuffer[1] != 0x82))
      printf("GENERAL_AUTHENTICATE failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FALSE;
  }

  return RC_TRUE;
}

/**f* PCSC/SCardReadBinary
 *
 * NAME
 *   SCardReadBinary
 *
 * DESCRIPTION
 *   Read data from a contactless card.
 *
 * INPUTS
 *   SCARDHANDLE hCard       : handle to the PC/SC smartcard
 *   BYTE address            : start address
 *   BYTE data[]             : buffer to receive the data
 *   BYTE size               : size of buffer
 *
 * OUTPUTS
 *   OK or error
 *
 **/
RC_T SCardReadBinary(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (prefer_specific)
    return SCardMifareClassicRead_Auto(hCard, address, data, size);
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xB0;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = size;

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 5, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(READ_BINARY) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if ((pcbRecvLength != (DWORD) (2+size)) && (pcbRecvLength != 2))
  {
    printf("SCardTransmit(READ_BINARY) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pcbRecvLength == 2) || (pbRecvBuffer[pcbRecvLength-2] != 0x90) || (pbRecvBuffer[pcbRecvLength-1] != 0x00))
  {
    if ((pbRecvBuffer[pcbRecvLength-2] == 0x6F) && (pbRecvBuffer[pcbRecvLength-1] == 0x0A))
    {
      /* Version 1-48 or earlier bug */
      if (SCardMifareClassicRead_Auto(hCard, address, data, size) == RC_TRUE)
      {
        prefer_specific = TRUE;
        return RC_TRUE;
      }
    }

    printf("READ_BINARY failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FATAL_ERROR;
  }

  if (data != NULL)
    memcpy(data, pbRecvBuffer, size);

  return RC_TRUE;
}

/**f* PCSC/SCardUpdateBinary
 *
 * NAME
 *   SCardUpdateBinary
 *
 * DESCRIPTION
 *   Write data to a contactless card.
 *
 * INPUTS
 *   SCARDHANDLE hCard       : handle to the PC/SC smartcard
 *   BYTE address            : start address
 *   BYTE data[]             : buffer holding the data
 *   BYTE size               : size of buffer
 *
 * OUTPUTS
 *   OK or error
 *
 **/
RC_T SCardUpdateBinary(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (prefer_specific)
    return SCardMifareClassicWrite_Auto(hCard, address, data, size);

  if (data == NULL) return RC_PARAMETER_ERROR;
  if (size + 5 > APDU_BUFFERS_SZ) return RC_PARAMETER_ERROR;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xD6;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = size;

  memcpy(&pbSendBuffer[5], data, size);

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 5+size, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(UPDATE_BINARY) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if (pcbRecvLength != 2)
  {
    printf("SCardTransmit(UPDATE_BINARY) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pbRecvBuffer[0] != 0x90) || (pbRecvBuffer[1] != 0x00))
  {
    if ((pbRecvBuffer[pcbRecvLength-2] == 0x6F) && (pbRecvBuffer[pcbRecvLength-1] == 0x0F))
    {
      /* Version 1-48 or earlier bug */
      if (SCardMifareClassicWrite_Auto(hCard, address, data, size) == RC_TRUE)
      {
        prefer_specific = TRUE;
        return RC_TRUE;
      }
    }

    printf("UPDATE_BINARY failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FALSE;
  }

  return RC_TRUE;
}


RC_T SCardMifareClassicRead_Auto(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xF3;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = size;

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 5, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(MIFARE_CLASSIC_READ) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if ((pcbRecvLength != (DWORD) (2+size)) && (pcbRecvLength != 2))
  {
    printf("SCardTransmit(MIFARE_CLASSIC_READ) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pcbRecvLength == 2) || (pbRecvBuffer[pcbRecvLength-2] != 0x90) || (pbRecvBuffer[pcbRecvLength-1] != 0x00))
  {
    printf("MIFARE_CLASSIC_READ failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FATAL_ERROR;
  }

  if (data != NULL)
    memcpy(data, pbRecvBuffer, size);

  return RC_TRUE;
}

RC_T SCardMifareClassicRead_KeyIndex(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size, BYTE key_location, BYTE key_index)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xF3;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = 2;
  pbSendBuffer[5] = key_location;
  pbSendBuffer[6] = key_index;
  pbSendBuffer[7] = size;

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 8, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(MIFARE_CLASSIC_READ) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if ((pcbRecvLength != (DWORD) (2+size)) && (pcbRecvLength != 2))
  {
    printf("SCardTransmit(MIFARE_CLASSIC_READ) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pcbRecvLength == 2) || (pbRecvBuffer[pcbRecvLength-2] != 0x90) || (pbRecvBuffer[pcbRecvLength-1] != 0x00))
  {
    if ((pbRecvBuffer[0] == 0x67) && (pbRecvBuffer[1] == 0x00))
    {
      /* Acceptable value, old versions don't support this */
      return RC_ACCEPTABLE_ERROR;
    }
    printf("MIFARE_CLASSIC_READ failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FATAL_ERROR;
  }

  if (data != NULL)
    memcpy(data, pbRecvBuffer, size);

  return RC_TRUE;
}

RC_T SCardMifareClassicRead_KeyValue(SCARDHANDLE hCard, BYTE address, BYTE data[], BYTE size, const BYTE key_value[6])
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (key_value == NULL) return RC_PARAMETER_ERROR;
  
  pbSendBuffer[0]  = 0xFF;
  pbSendBuffer[1]  = 0xF3;
  pbSendBuffer[2]  = 0x00;
  pbSendBuffer[3]  = address;
  pbSendBuffer[4]  = 6;
  memcpy(&pbSendBuffer[5], key_value, 6);
  pbSendBuffer[11] = size;

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 12, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(MIFARE_CLASSIC_READ) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if ((pcbRecvLength != (DWORD) (2+size)) && (pcbRecvLength != 2))
  {
    printf("SCardTransmit(MIFARE_CLASSIC_READ) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pcbRecvLength == 2) || (pbRecvBuffer[pcbRecvLength-2] != 0x90) || (pbRecvBuffer[pcbRecvLength-1] != 0x00))
  {
    printf("MIFARE_CLASSIC_READ failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FATAL_ERROR;
  }

  if (data != NULL)
    memcpy(data, pbRecvBuffer, size);

  return RC_TRUE;
}

RC_T SCardMifareClassicWrite_Auto(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (data == NULL) return RC_PARAMETER_ERROR;
  if (size + 5 > APDU_BUFFERS_SZ) return RC_PARAMETER_ERROR;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xF4;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = size;

  memcpy(&pbSendBuffer[5], data, size);

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 5+size, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(MIFARE_CLASSIC_WRITE) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if (pcbRecvLength != 2)
  {
    printf("SCardTransmit(MIFARE_CLASSIC_WRITE) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pbRecvBuffer[0] != 0x90) || (pbRecvBuffer[1] != 0x00))
  {
    printf("MIFARE_CLASSIC_WRITE failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FALSE;
  }

  return RC_TRUE;
}

RC_T SCardMifareClassicWrite_KeyIndex(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size, BYTE key_location, BYTE key_index)
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (data == NULL) return RC_PARAMETER_ERROR;
  if (size + 5 + 2 > APDU_BUFFERS_SZ) return RC_PARAMETER_ERROR;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xF4;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = size + 2;

  memcpy(&pbSendBuffer[5], data, size);
  pbSendBuffer[5 + size] = key_location;
  pbSendBuffer[6 + size] = key_index;

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 7+size, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(MIFARE_CLASSIC_WRITE) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if (pcbRecvLength != 2)
  {
    printf("SCardTransmit(MIFARE_CLASSIC_WRITE) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pbRecvBuffer[0] != 0x90) || (pbRecvBuffer[1] != 0x00))
  {
    if ((pbRecvBuffer[0] == 0x67) && (pbRecvBuffer[1] == 0x00))
    {
      /* Acceptable value, old versions don't support this */
      return RC_ACCEPTABLE_ERROR;
    }
    printf("MIFARE_CLASSIC_WRITE failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FALSE;
  }

  return RC_TRUE;
}

RC_T SCardMifareClassicWrite_KeyValue(SCARDHANDLE hCard, BYTE address, const BYTE data[], BYTE size, const BYTE key_value[6])
{
  BYTE pbSendBuffer[APDU_BUFFERS_SZ];
  BYTE  pbRecvBuffer[APDU_BUFFERS_SZ];
  DWORD pcbRecvLength = sizeof(pbRecvBuffer);

  LONG rc;

  if (data == NULL) return RC_PARAMETER_ERROR;
  if (key_value == NULL) return RC_PARAMETER_ERROR;
  if (size + 5 + 6 > APDU_BUFFERS_SZ) return RC_PARAMETER_ERROR;
  
  pbSendBuffer[0] = 0xFF;
  pbSendBuffer[1] = 0xF4;
  pbSendBuffer[2] = 0x00;
  pbSendBuffer[3] = address;
  pbSendBuffer[4] = size + 6;

  memcpy(&pbSendBuffer[5], data, size);
  memcpy(&pbSendBuffer[5 + size], key_value, 6);

  rc = SCardTransmit(hCard, SCARD_PCI_T1, pbSendBuffer, 5+size+6, NULL, pbRecvBuffer, &pcbRecvLength);
  if (rc != SCARD_S_SUCCESS)
	{
    printf("SCardTransmit(MIFARE_CLASSIC_WRITE) error %08lX\n", rc);
		return RC_FATAL_ERROR;
	}

  if (pcbRecvLength != 2)
  {
    printf("SCardTransmit(MIFARE_CLASSIC_WRITE) returned %ld bytes\n", pcbRecvLength);
    return RC_FATAL_ERROR;
  }

  if ((pbRecvBuffer[0] != 0x90) || (pbRecvBuffer[1] != 0x00))
  {
    printf("MIFARE_CLASSIC_WRITE failed, SW=%02X%02X\n", pbRecvBuffer[0], pbRecvBuffer[1]);
    return RC_FALSE;
  }

  return RC_TRUE;
}

/* Mifare transport condition */
const BYTE AccessTransport[4] = { 0x00, 0x00, 0x00, 0x01 };
/* Mifare usage condition (with data blocks, not counters) */
const BYTE AccessUsage[4]     = { 0x04, 0x04, 0x04, 0x03 };

/* Default Mifare transport key */
const BYTE KeyTransport[6]    = { 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF };
const BYTE KeyEmpty[6]        = { 0x00,0x00,0x00,0x00,0x00,0x00 };

/* The keys used for our tests */
const BYTE KeyTestA[6]        = { 0xAA,0xAA,0xAA,0xAA,0xAA,0xAA };
const BYTE KeyTestB[6]        = { 0xBB,0xBB,0xBB,0xBB,0xBB,0xBB };


/*
 * SCardLoadKeys
 * -------------
 * Load the keys we are to use into reader's memory
 */
RC_T SCardLoadKeys(SCARDHANDLE hCard)
{
  RC_T rc;
  
  rc = SCardLoadKey(hCard, KeyLocation, KeyTestAIdx, KeyTestA, sizeof(KeyTestA));
  if (rc != RC_TRUE)
  {
    printf("Failed to load test key A\n");
    return rc;
  }

  rc = SCardLoadKey(hCard, KeyLocation, KeyTestBIdx, KeyTestB, sizeof(KeyTestB));
  if (rc != RC_TRUE)
  {
    printf("Failed to load test key B\n");
    return rc;
  }

  rc = SCardLoadKey(hCard, KeyLocation, KeyTestTIdx, KeyTransport, sizeof(KeyTransport));
  if (rc != RC_TRUE)
  {
    printf("Failed to load transport key\n");
    return rc;
  }

  return RC_TRUE;
}

/*
 * SCardMifareIsTransport
 * ----------------------
 * Verify that the card is in transport condition (KeyTransport OK on every sector,
 * all data are 0x00, writing is possible with KeyTransport)
 */
RC_T SCardMifareIsTransport(SCARDHANDLE hCard)
{
  BYTE data[3*16]; /* 3 blocks */
  BYTE sector, i;
  RC_T rc;

  for (sector=0; sector<16; sector++)
  {
    /* KeyTransport must be OK */
    rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, KeyTestTIdx);
    if (rc == RC_FALSE)
    {
      printf("Sector %d : authentication with transport key failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR during authentication with transport key\n", sector);
      return rc;
    }

    /* Read 3 blocks */
    rc = SCardReadBinary(hCard, (BYTE) (4 * sector), data, sizeof(data));
    if (rc == RC_FALSE)
    {
      printf("Sector %d : read with transport key failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR while reading with transport key\n", sector);
      return rc;
    }

    /* Check that the blocks are all blanks */
    for (i=sector?0:16; i<sizeof(data); i++)
    {      
      if (data[i] != 0x00)
      {
        printf("data[%d] = %02X\n", i, data[i]); 
        break;
      } 
    }

    if (i < sizeof(data))
    {
      printf("Sector %d is not blank\n", sector);
      return RC_FALSE;
    }

    /* Re-write (1 block is enough and faster, don't write block 0 of sector 0) */
    rc = SCardUpdateBinary(hCard, (BYTE) (4 * sector + 1), &data[16], 16);
    if (rc == RC_FALSE)
    {
      printf("Sector %d : write with transport key failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR while writing with transport key\n", sector);
      return rc;
    }
  }

  /* OK, card is actually virgin ! */
  return RC_TRUE;
}

/*
 * SCardMifareFormatTransport
 * --------------------------
 * Put back the card in transport condition
 */
RC_T SCardMifareFormatTransport(SCARDHANDLE hCard)
{
  BYTE data[3*16]; /* 3 blocks */
  BYTE trailer[16];
  BYTE sector;
  RC_T rc;

  /* We'll write blank data */
  memset(data, 0x00, sizeof(data));

  /* We'll format every sector this way : */
  MakeMifareSectorTrailer(trailer, KeyTransport, KeyEmpty, AccessTransport);

  for (sector=0; sector<16; sector++)
  {
    /* Let's get authenticated */
    rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, KeyTestTIdx);
    if (rc == RC_FALSE)
    {
      /* Failed to get authenticated with transport key, let's try our B test key */
      rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, KeyTestBIdx);
      if (rc == RC_FALSE)
      {
        printf("Sector %d : authentication failed, don't know the key !\n", sector);
        return rc;
      }
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR during authentication\n", sector);
      return rc;
    }

    /* Write blank data */
    rc = SCardUpdateBinary(hCard, (BYTE) (4 * sector), data, 3*16);
    if (rc == RC_FALSE)
    {
      printf("Sector %d : write blank data failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR while writing blank data\n", sector);
      return rc;
    }

    /* Write the transport key (+ transport access conditions) in sector trailer */
    rc = SCardUpdateBinary(hCard, (BYTE) (4 * sector + 3), trailer, 16);
    if (rc == RC_FALSE)
    {
      printf("Sector %d : write sector trailer failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR while writing sector trailer\n", sector);
      return rc;
    }
  }

  /* OK */
  return RC_TRUE;
}

/*
 * SCardMifareFormatUsage
 * ----------------------
 * Format the card with the key A and key B that will be used during the tests
 */
RC_T SCardMifareFormatUsage(SCARDHANDLE hCard)
{
  BYTE data[3*16]; /* 3 blocks */
  BYTE trailer[16];
  BYTE sector;
  RC_T rc;

  /* We'll write blank data */
  memset(data, 0x00, sizeof(data));

  /* We'll format every sector this way : */
  MakeMifareSectorTrailer(trailer, KeyTestA, KeyTestB, AccessUsage);

  for (sector=0; sector<16; sector++)
  {
    /* Let's get authenticated */
    rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, KeyTestTIdx);
    if (rc == RC_FALSE)
    {
      /* Failed to get authenticated with transport key, let's try our B test key */
      rc = SCardGeneralAuthenticate(hCard, (BYTE) (4 * sector), KeyLocation, KeyTestBIdx);
      if (rc == RC_FALSE)
      {
        printf("Sector %d : authentication failed, don't know the key !\n", sector);
        return rc;
      }
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR during authentication\n", sector);
      return rc;
    }

    /* Write blank data */
    rc = SCardUpdateBinary(hCard, (BYTE) (4 * sector), data, 3*16);
    if (rc == RC_FALSE)
    {
      printf("Sector %d : write blank data failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR while writing blank data\n", sector);
      return rc;
    }

    /* Write the transport key (+ transport access conditions) in sector trailer */
    rc = SCardUpdateBinary(hCard, (BYTE) (4 * sector + 3), trailer, 16);
    if (rc == RC_FALSE)
    {
      printf("Sector %d : write sector trailer failed\n", sector);
      return rc;
    }
    if (rc != RC_TRUE)
    {
      printf("Sector %d : ERROR while writing sector trailer\n", sector);
      return rc;
    }
  }

  /* OK */
  return RC_TRUE;
}


#define Bit2(a) ((a&0x04)?1:0)
#define Bit1(a) ((a&0x02)?1:0)
#define Bit0(a) ((a&0x01)?1:0)

void MakeMifareSectorTrailer(BYTE trailer[16], const BYTE key_a[6], const BYTE key_b[6], const BYTE ac[4])
{
  BYTE C3, C2, C1;

  if (trailer == NULL) return;

  memset(trailer, 0xFF, 16);

  /* Put the 2 keys */
  if (key_a != NULL)
    memcpy(&trailer[0], key_a, 6);

  if (key_b != NULL)
    memcpy(&trailer[10], key_b, 6);

  /* Build the access conditions group */  
  C3  = Bit0(ac[3]);
  C3 <<= 1;
  C3 |= Bit0(ac[2]);
  C3 <<= 1;
  C3 |= Bit0(ac[1]);
  C3 <<= 1;
  C3 |= Bit0(ac[0]);

  C2  = Bit1(ac[3]);
  C2 <<= 1;
  C2 |= Bit1(ac[2]);
  C2 <<= 1;
  C2 |= Bit1(ac[1]);
  C2 <<= 1;
  C2 |= Bit1(ac[0]);

  C1  = Bit2(ac[3]);
  C1 <<= 1;
  C1 |= Bit2(ac[2]);
  C1 <<= 1;
  C1 |= Bit2(ac[1]);
  C1 <<= 1;
  C1 |= Bit2(ac[0]);

  trailer[6] = (~(C2 << 4) & 0xF0) | (~(C1) & 0x0F);
  trailer[7] = ((C1 << 4) & 0xF0) | (~(C3) & 0x0F);
  trailer[8] = ((C3 << 4) & 0xF0) | (C2 & 0x0F);
  trailer[9] = 0x00;
}


#ifdef WIN32

static LARGE_INTEGER Before = { 0 };

void StartInterval(void)
{
  QueryPerformanceCounter(&Before);
}

DWORD GetInterval(void)
{
  DWORD r;
  LARGE_INTEGER Now, Freq;

  QueryPerformanceCounter(&Now);

  QueryPerformanceFrequency(&Freq);
  Freq.QuadPart /= 1000;

  r = (DWORD) ( (Now.QuadPart - Before.QuadPart) / Freq.QuadPart );
  return r;
}

#endif
