/**
 *
 * NAME
 *   atr_analysis.java :: Explain an ISO 7816-3 ATR (C implementation)
 *
 * DESCRIPTION
 *   The ATR_analysis function decodes a smartcard ATR, an explain its
 *   fields.
 *   The ATR_identify function finds and display information from the
 *   smartcard_list.txt file.
 *
 * COPYRIGHT
 *   Copyright (c) 2009 Johann Dantant - Emilie Clement / Pro-Active
 *   Thanks Ludovic Rousseau, Christophe Levantis for the skeleton of the ATR_analysis
 *   function (the function is a port of their PERL script to C)
 *
 * LICENSE
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **/
public class ATR_Analysis {
	/*
	 * For more information about the ATR see ISO 7816-3 1997, pages 12 and up
	 * 
	 * TS	initial character
	 * T0	format character, Y1 | K
	 * 	interfaces characters
	 * TA(1)	global, codes  FI and DI
	 * TB(1)	global, codes II and PI
	 * TC(1)	global, codes N
	 * TD(1)	codes Y(2) and T
	 * TA(2)	global, codes specific mode
	 * TB(2)	global, codes PI2
	 * TC(2)	specific
	 * TD(2)	codes Y(3) and T
	 * TA(3)	TA(i), TB(i), TC(i) (i>2)
	 * 		- specific after T!=15
	 * 		- global after T=15
	 * TD(i)	codes Y(i+1) and T
	 * T1	historical bytes
	 * 		max 15 characters
	 * TK
	 * TCK	check character
	 */	
	
	static String BINQ[] = {
	  "0000","0001","0010","0011","0100","0101","0110","0111",
	  "1000","1001","1010","1011","1100","1101","1110","1111"
	};
	
	static String Fi[]= 
	{
	  "372", "372", "558",  "744", "1116", "1488", "1860", "RFU",
	  "RFU", "512", "768", "1024", "1536", "2048", "RFU",  "RFU"
	};

	static String Di[] =
	{
	  "RFU", "1",   "2",   "4",   "8",   "16",  "32", "RFU",
	  "12", "20", "RFU", "RFU", "RFU",	"RFU", "RFU", "RFU"
	};
	
	/*     _   _____  _____
	 *    / \ |_   _||  _  \
	 *   / _ \  | |  | |_| |
	 *  / ___ \ | |  |  __ \
	 * /_/   \_\|_|  |_|  \_\
	 */
	public ATR_Analysis(byte atr[], int atrlen)
	{ 
	  int offset = 0;

	  byte Y1;
	  byte K;
     
	  /* Display ATR */
	  System.out.print("ATR :");
	  for(int i=0;i<atr.length;i++)
		  System.out.print(String.format(" %02X" , atr[i]));
	  System.out.print("\n");

	  /*
	   * Explain ATR content
	   * -------------------
	   */

	  /* TS */
	  /* -- */
	  System.out.print("\t\t+ TS  = "+String.format("%02X --> ",atr[0]));
	  switch (atr[0])
	  {
	    case 0x3B : System.out.print("Direct Convention");  break;
	    case 0x3F : System.out.print("Inverse Convention"); break;
	    default   : System.out.print("UNDEFINED"); break;
	  }
	  System.out.print("\n");
	  /* T0 */
	  /* -- */
	  Y1 = (byte) ((atr[1]&0xFF) / 0x10);
	  K  = (byte) ((atr[1]&0xFF) % 0x10);
	  
	  System.out.print("\t\t+ T0  = "+String.format("%02X",atr[1])+", Y1="+BINQ[Y1]+", K="+String.format("%d", K)+" (historical bytes)\n");

	  /* TA, TB, TC, TD */
	  /* -------------- */

	  offset = 2;

	  if ((Y1 & 0x01)>0)
	    ATR_analysis_TA((byte)1,(byte) 0, offset, atr, atrlen);

	  if ((Y1 & 0x02)>0)
	    ATR_analysis_TB((byte)1,(byte) 0, offset, atr, atrlen);

	  if ((Y1 & 0x04)>0)
	    ATR_analysis_TC((byte)1,(byte) 0, offset, atr, atrlen);

	  if ((Y1 & 0x08) >0)
	    ATR_analysis_TD((byte)1, offset, atr, atrlen);

	  /* Historical bytes */
	  /* ---------------- */
	  if (K>0)
	  {
	    int i;
	    String historicalByte = K>1?"s":"";
	    System.out.print("\t\t+ "+String.format("%d", K)+" historical byte"+historicalByte +":\n");
	    System.out.print("\t\t       ");

	    for (i=0; i<K; i++)
	    {
	      if (offset >= atrlen)
	      {
	    	  System.out.print("\n\t\tERROR! ATR is truncated ("+String.format("%d", (K-i))+" bytes missing)\n");
	        break;
	      }
	      System.out.print(" "+ String.format("%02X", atr[offset++]));
	    }
	    System.out.print("\n");
	  }

	  /* Check TCK */
	  /* --------- */
	  if (K>0)
	  {
	    int i;
	    byte  TCK = 0;
	    
	    for (i=1; i<(atrlen-1); i++)
	    {
	      TCK ^= atr[i];
	    }

	    System.out.print("\t\t+ TCK = "+String.format("%02X", TCK));

	    if (TCK == atr[atrlen-1])
	    {
	      System.out.print("(correct checksum)");
	    } else
	    {
	      System.out.print("!= "+ String.format("%02X", atr[atrlen-1]));
	      System.out.print("\n\t\tERROR! Checksum is invalid\n");
	    }
	    System.out.print("\n");
	  }

	  System.out.print("\n");
	}
	
	/*  _____  _    
	 * |_   _|/ \   
	 *   | | / _ \  
	 *   | |/ ___ \ 
	 *   |_/_/   \_\
	 */
	void ATR_analysis_TA(byte counter, byte proto, int offset, byte atr[], int atrlen)
	{
	  byte value;
	 
	  value = atr[offset];
	  offset = offset + 1;

      System.out.print("\t\t+ TA"+String.format("%d", counter)+" = "+String.format("%02X", value)+" --> ");

	  if (counter == 1)
	  {
	    /* TA1 */
	    byte F = (byte) ((value&0xFF) / 0x10);
	    byte D = (byte) ((value&0xFF) % 0x10);

	    System.out.print("Fi="+Fi[F]+", Di="+Di[D]);

	  } else
	  if (counter == 2)
	  {
	    /* TA2 */
	    byte F = (byte) ((value&0xFF)/ 0x10);
	    byte D = (byte) ((value&0xFF) % 0x10);

	    System.out.print("Protocol to be used in spec mode: T="+String.format("%d", D));
		if ( (F & 0x08) >0)
		{
			System.out.print(" - unable to change");
		} else
		{
			System.out.print(" - capable to change");
		}
		if ( (F & 0x01) >0)
		{
			System.out.print(" - implicity defined");
		}
		else
		{
			System.out.print(" - defined by interface bytes");
		}
	  } else
	  if (counter >= 3)
	  {
		  /* TA3 (and other) */
	    if (proto == 1)
	    {
	      /* T=1 protocol */
	      System.out.print("IFSC: "+String.format("%d", value));
		  } else
		  {
	      /* Other protocol than T=1 */
	      byte F = (byte) ((value&0xFF) / 0x40);
	      byte D = (byte) ((value&0xFF) % 0x40);

	      System.out.print("Class: ");

	      if ( (D & 0x01)>0)
	    	System.out.print("A 5V ");
	      if ( (D & 0x02)>0)
	    	System.out.print("B 3V ");
	      if ( (D & 0x04)>0)
	    	System.out.print("C 1.8V ");
	      if ( (D & 0x08)>0)
	    	System.out.print("D RFU ");
	      if ( (D & 0x10)>0)
	    	System.out.print("E RFU ");

	      System.out.print("Clock stop: ");
	      switch (F)
	      {
	        case 0 : System.out.print("not supported"); break;
	        case 1 : System.out.print("state L"); break;
	        case 2 : System.out.print("state H"); break;
	        case 3 : System.out.print("no preference");
	      }
	    }
	  }

	  System.out.print("\n");
	}

	/*  _____ ____  
	 * |_   _| __ ) 
	 *   | | |  _ \ 
	 *   | | | |_) |
	 *   |_| |____/ 
	 */
	void ATR_analysis_TB(byte counter, byte T, int offset, byte atr[], int atrlen)
	{
	  byte value;
	 
	  value = atr[offset];
	  offset = offset + 1;

	  System.out.print("\t\t+ TB"+String.format("%d", counter)+" = "+String.format("%02X", value)+" --> ");


	  if (counter == 1)
	  {
	    /* TB1 */
	    byte I  = (byte) ((value&0xFF) / 0x20);
	    byte PI = (byte) ((value&0xFF) % 0x20);

	    if (PI == 0)
	    {
	      System.out.print("Vpp not connected");
	    } else
	    {
	      System.out.print("Programming params P:"+String.format("%d", PI)+"V, I:"+String.format("%d", I)+"mA");
	    }
	  } else
	  if (counter == 2)
	  {
	    /* TB2 */
		  System.out.println("Programming param PI2 (PI1 should be ignored): ");
			if ((value>49) || (value<251))
			{
				System.out.print(String.format("%d (dV)", value));
			}
			else
			{
				System.out.print(String.format("RFU value %d", value));
			}
	  } else
	  if (counter >= 3)
	  {
	    if (T == 1)
	    {
	      byte BWI = (byte) ((value&0xFF) / 0x10);
	      byte CWI = (byte) ((value&0xFF) % 0x10);

	      System.out.print("BWI="+String.format("%d", BWI)+" - CWI="+String.format("%d", CWI));
	    }
	  }



	  System.out.print("\n");
	}
	
	/*
	 *  _____ ____ 
	 * |_   _/ ___|
	 *   | || |    
	 *   | || |___ 
	 *   |_| \____|
	 */
	void ATR_analysis_TC(byte counter, byte T, int offset, byte atr[], int atrlen)
	{
	  byte value;
	 
	  value = atr[offset];
	  offset = offset + 1;

	  System.out.print("\t\t+ TC"+String.format("%d", counter)+" = "+String.format("%02X", value)+" --> ");

	  if (counter == 1)
	  {
	    /* TC1 */
		System.out.print(String.format("EGT=%d", value));
	    if (value == 0xFF)
	      System.out.print(" (special value)");
	  } else
	  if (counter == 2)
	  {
	    /* TC2 */
		System.out.print("Work waiting time= 960 x "+String.format("%d", value)+" x (Fi/F)");
	  } else
	  if (counter == 3)
	  {
	    /* TC3 */
	    if (T == 1)
	    {
	      System.out.print("Error dectection code: ");
	      if (value == 1)
	      {
	    	System.out.print("CRC");
	      } else
	      if (value == 0)
	      {
	    	System.out.print("LRC");
	      } else
	      {
	    	System.out.print("RFU");
	      }
	    }
	  }

	  System.out.print("\n");
	}
	
	/*  _____ ____  
	 * |_   _|  _ \ 
	 *   | | | | | |
	 *   | | | |_| |
	 *   |_| |____/ 
	 */
	void ATR_analysis_TD(byte counter, int offset, byte atr[], int atrlen)
	{
	  byte value, Y, T;
	 
	  value = atr[offset];
	  offset = offset + 1;

	  Y = (byte) ((value&0xFF) / 0x10);
	  T = (byte) ((value&0xFF) % 0x10);

	  if (T == 15)
	  {
		System.out.print("\t\tGlobal interface bytes following\n");
	  }

	  System.out.println("\t\t+ TD"+String.format("%d", counter)+" = "+String.format("%02X", value)+"  --> Y"+String.format("%d", (counter+1))+"="+BINQ[Y]+", protocol T="+String.format("%d", T)+"\n");
	  System.out.print("\t\t-----------\n");

	  counter++;

	  if ((Y & 0x01)>0)
	    ATR_analysis_TA(counter, T, offset, atr, atrlen);

	  if ((Y & 0x02)>0)
	    ATR_analysis_TB(counter, T, offset, atr, atrlen);

	  if ((Y & 0x04)>0)
	    ATR_analysis_TC(counter, T, offset, atr, atrlen);

	  if ((Y & 0x08) >0)
	    ATR_analysis_TD(counter, offset, atr, atrlen);

	}

}
